from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json

from compas.robots import ToolModel, Geometry

from compas_fab.robots.planning_scene import AttachedCollisionMesh
from compas_fab.robots.planning_scene import CollisionMesh


class Tool(object):
    """Represents a tool to be attached to the robot's flange.

    Attributes
    ----------
    visual : :class:`compas.datastructures.Mesh`
        The visual mesh of the tool.
    frame_in_tool0_frame : :class:`compas.geometry.Frame`
        The frame of the tool in tool0 frame.
    collision : :class:`compas.datastructures.Mesh`
        The collision mesh representation of the tool.
    name : :obj:`str`
        The name of the `Tool`. Defaults to 'attached_tool'.
    link_name : :obj:`str`
        The name of the `Link` to which the tool is attached.  Defaults to ``None``.

    Examples
    --------
    >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/cone.stl'))
    >>> frame = Frame([0.14, 0, 0], [0, 1, 0], [0, 0, 1])
    >>> tool = Tool(mesh, frame)

    """

    def __init__(self, visual, frame_in_tool0_frame, collision=None,
                 name="attached_tool", link_name=None):
        self.tool_model = ToolModel(visual, frame_in_tool0_frame, collision, name, link_name)

    @classmethod
    def from_tool_model(cls, tool_model):
        tool = cls(None, None)
        tool.tool_model = tool_model
        return tool

    @property
    def attached_collision_meshes(self):
        acms = []
        for link in self.tool_model.iter_links():
            for i, item in enumerate(link.collision):
                meshes = Geometry._get_item_meshes(item)
                for mesh in meshes:
                    collision_mesh_name = '{}_collision_{}'.format(link.name, i)
                    collision_mesh = CollisionMesh(mesh, collision_mesh_name)
                    attached_collision_mesh = AttachedCollisionMesh(collision_mesh, self.link_name, [self.link_name])
                    acms.append(attached_collision_mesh)
        return acms

    @property
    def name(self):
        return self.tool_model.name

    @property
    def link_name(self):
        return self.tool_model.link_name

    @link_name.setter
    def link_name(self, link_name):
        self.tool_model.link_name = link_name

    @property
    def frame(self):
        return self.tool_model.frame

    @frame.setter
    def frame(self, frame):
        self.tool_model.frame = frame

    @property
    def data(self):
        """Returns the data dictionary that represents the tool.

        Returns
        -------
        :obj:`dict`
            The frame data.

        """
        data = self.tool_model.data
        return data

    @data.setter
    def data(self, data):
        self.tool_model.data = data

    @classmethod
    def from_data(cls, data):
        """Construct a `Tool` from its data representation.

        Parameters
        ----------
        data : :obj:`dict`
            The data dictionary.

        Returns
        -------
        :class:`Tool`
            The constructed `Tool`.

        Examples
        --------
        >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/cone.stl'))
        >>> frame = Frame([0.14, 0, 0], [0, 1, 0], [0, 0, 1])
        >>> data = {'visual': mesh.data, 'frame': frame.data}
        >>> tool = Tool.from_data(data)
        """
        tool = cls(None, None)
        tool.data = data
        return tool

    @classmethod
    def from_json(cls, filepath):
        """Construct a `Tool` from the data contained in a JSON file.

        Parameters
        ----------
        filepath : str
            Path to the file containing the data.

        Returns
        -------
        :class:`Tool`
            The tool.

        Examples
        --------
        >>> filepath = os.path.join(compas_fab.DATA, "planning_scene", "cone_tool.json")
        >>> tool = Tool.from_json(filepath)
        """
        with open(filepath, 'r') as f:
            data = json.load(f)
        return cls.from_data(data)

    def to_json(self, filepath):
        """Serialise the data dictionary representing the tool to JSON and store in a file.

        Parameters
        ----------
        filepath : :obj:`str`
            Path to the file.

        Returns
        -------
        None

        Examples
        --------
        >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/cone.stl'))
        >>> frame = Frame([0.14, 0, 0], [0, 1, 0], [0, 0, 1])
        >>> tool = Tool(mesh, frame)
        >>> filepath = os.path.join(compas_fab.DATA, "planning_scene", "cone_tool.json")
        >>> tool.to_json(filepath)
        """
        with open(filepath, 'w') as f:
            json.dump(self.data, f, indent=4, sort_keys=True)

    def update_touch_links(self, touch_links=None):
        if touch_links:
            for acm in self.attached_collision_meshes:
                acm.touch_links = touch_links

    def from_tcf_to_t0cf(self, frames_tcf):
        """Converts a list of frames at the robot's tool tip (tcf frame) to frames at the robot's flange (tool0 frame).

        Parameters
        ----------
        frames_tcf : :obj:`list` of :class:`compas.geometry.Frame`
            Frames (in WCF) at the robot's tool tip (tcf).

        Returns
        -------
        :obj:`list` of :class:`compas.geometry.Frame`
            Frames (in WCF) at the robot's flange (tool0).

        Examples
        --------
        >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/cone.stl'))
        >>> frame = Frame([0.14, 0, 0], [0, 1, 0], [0, 0, 1])
        >>> tool = Tool(mesh, frame)
        >>> frames_tcf = [Frame((-0.309, -0.046, -0.266), (0.276, 0.926, -0.256), (0.879, -0.136, 0.456))]
        >>> tool.from_tcf_to_t0cf(frames_tcf)
        [Frame(Point(-0.363, 0.003, -0.147), Vector(0.388, -0.351, -0.852), Vector(0.276, 0.926, -0.256))]
        """
        return self.tool_model.from_tcf_to_t0cf(frames_tcf)

    def from_t0cf_to_tcf(self, frames_t0cf):
        """Converts frames at the robot's flange (tool0 frame) to frames at the robot's tool tip (tcf frame).

        Parameters
        ----------
        frames_t0cf : :obj:`list` of :class:`compas.geometry.Frame`
            Frames (in WCF) at the robot's flange (tool0).

        Returns
        -------
        :obj:`list` of :class:`compas.geometry.Frame`
            Frames (in WCF) at the robot's tool tip (tcf).

        Examples
        --------
        >>> mesh = Mesh.from_stl(compas_fab.get('planning_scene/cone.stl'))
        >>> frame = Frame([0.14, 0, 0], [0, 1, 0], [0, 0, 1])
        >>> tool = Tool(mesh, frame)
        >>> frames_t0cf = [Frame((-0.363, 0.003, -0.147), (0.388, -0.351, -0.852), (0.276, 0.926, -0.256))]
        >>> tool.from_t0cf_to_tcf(frames_t0cf)
        [Frame(Point(-0.309, -0.046, -0.266), Vector(0.276, 0.926, -0.256), Vector(0.879, -0.136, 0.456))]
        """
        return self.tool_model.from_t0cf_to_tcf(frames_t0cf)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__all__ = [
    'Duration',
]


class Duration(object):
    """Duration consists of two integers: seconds and nanoseconds.

    Attributes
    ----------
    secs: int
        Integer representing number of seconds.
    nsecs: int
        Integer representing number of nanoseconds.
    """

    def __init__(self, secs, nsecs):
        self.secs = int(secs)
        self.nsecs = int(nsecs)

    def __str__(self):
        return 'Duration({}, {})'.format(self.secs, self.nsecs)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, Duration):
            return self.seconds == other.seconds

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def seconds(self):
        """Returns the duration as floating-point seconds.

        Returns
        -------
        float
            Floating-point seconds
        """
        return self.secs + 1e-9 * self.nsecs

    @classmethod
    def from_data(cls, data):
        """Construct a duration from its data representation.

        Parameters
        ----------
        data : :obj:`dict`
            The data dictionary.

        Returns
        -------
        :class:`Duration`
             An instance of :class:`Duration`.
        """
        duration = cls(0, 0)
        duration.data = data
        return duration

    def to_data(self):
        """Return the data dictionary that represents the duration, and from
        which it can be reconstructed."""
        return self.data

    @property
    def data(self):
        """:obj:`dict` : The data representing the duration."""
        return {
            'secs': self.secs,
            'nsecs': self.nsecs
        }

    @data.setter
    def data(self, data):
        self.secs = data.get('secs') or 0
        self.nsecs = data.get('nsecs') or 0



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import namedtuple

PYBULLET_GUI = 1
PYBULLET_DIRECT = 2
PYBULLET_SHARED_MEMORY = 3
PYBULLET_UDP = 4
PYBULLET_TCP = 5
PYBULLET_GUI_SERVER = 7
PYBULLET_SHARED_MEMORY_SERVER = 9
PYBULLET_SHARED_MEMORY_GUI = 14

CONNECTION_TYPE = {
    'direct': PYBULLET_DIRECT,
    'gui': PYBULLET_GUI,
    'shared_memory': PYBULLET_SHARED_MEMORY,
    'udp': PYBULLET_UDP,
    'tcp': PYBULLET_TCP,
    'gui_server': PYBULLET_GUI_SERVER,
    'shared_memory_server': PYBULLET_SHARED_MEMORY_SERVER,
    'shared_memory_gui': PYBULLET_SHARED_MEMORY_GUI,
}

BASE_LINK_ID = -1
NULL_ID = -1

STATIC_MASS = 0

GREY = (0.5, 0.5, 0.5, 1)
RED = (1, 0, 0, 1)

BodyInfo = namedtuple('BodyInfo', ['base_name', 'body_name'])
ConstraintInfo = namedtuple('ConstraintInfo', ['constraint_id', 'body_id', 'robot_uid'])
JointInfo = namedtuple('JointInfo', ['jointIndex', 'jointName', 'jointType',
                                     'qIndex', 'uIndex', 'flags',
                                     'jointDamping', 'jointFriction', 'jointLowerLimit', 'jointUpperLimit',
                                     'jointMaxForce', 'jointMaxVelocity', 'linkName', 'jointAxis',
                                     'parentFramePos', 'parentFrameOrn', 'parentIndex'])
JointState = namedtuple('JointState', ['jointPosition', 'jointVelocity',
                                       'jointReactionForces', 'appliedJointMotorTorque'])
LinkState = namedtuple('LinkState', ['linkWorldPosition', 'linkWorldOrientation',
                                     'localInertialFramePosition', 'localInertialFrameOrientation',
                                     'worldLinkFramePosition', 'worldLinkFrameOrientation'])



"""
********************************************************************************
compas_fab.blender
********************************************************************************

.. currentmodule:: compas_fab.blender

Artists
-------

In **COMPAS**, the `artists` are classes that assist with the visualization of
datastructures and models, in a way that maintains the data separated from the
specific CAD interfaces, while providing a way to leverage native performance
of the CAD environment.

This package only provides a deprecated alias to the
:class:`~compas_blender.artists.RobotModelArtist` implemented in COMPAS.

"""
try:
    import bpy  # noqa: F401
    import mathutils  # noqa: F401
except ImportError:
    pass
else:
    from .artists import RobotArtist
    from .artists import RobotModelArtist

    __all__ = ['RobotArtist', 'RobotModelArtist']



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


try:
    import bpy  # noqa: F401
    import mathutils  # noqa: F401
except ImportError:
    pass
else:
    from compas_blender.artists import RobotModelArtist

    __all__ = [
        'RobotModelArtist',
    ]

    # deprecated alias
    RobotArtist = RobotModelArtist



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.interfaces.client import forward_docstring
from compas_fab.backends.interfaces.client import PlannerInterface
from compas_fab.backends.pybullet.backend_features.pybullet_add_attached_collision_mesh import PyBulletAddAttachedCollisionMesh
from compas_fab.backends.pybullet.backend_features.pybullet_forward_kinematics import PyBulletForwardKinematics
from compas_fab.backends.pybullet.backend_features.pybullet_inverse_kinematics import PyBulletInverseKinematics
from compas_fab.backends.pybullet.backend_features.pybullet_add_collision_mesh import PyBulletAddCollisionMesh
from compas_fab.backends.pybullet.backend_features.pybullet_append_collision_mesh import PyBulletAppendCollisionMesh
from compas_fab.backends.pybullet.backend_features.pybullet_remove_attached_collision_mesh import PyBulletRemoveAttachedCollisionMesh
from compas_fab.backends.pybullet.backend_features.pybullet_remove_collision_mesh import PyBulletRemoveCollisionMesh


class PyBulletPlanner(PlannerInterface):
    """Implement the planner backend interface for PyBullet."""
    def __init__(self, client):
        super(PyBulletPlanner, self).__init__(client)

    @forward_docstring(PyBulletAddAttachedCollisionMesh)
    def add_attached_collision_mesh(self, *args, **kwargs):
        return PyBulletAddAttachedCollisionMesh(self.client)(*args, **kwargs)

    @forward_docstring(PyBulletAddCollisionMesh)
    def add_collision_mesh(self, *args, **kwargs):
        return PyBulletAddCollisionMesh(self.client)(*args, **kwargs)

    @forward_docstring(PyBulletAppendCollisionMesh)
    def append_collision_mesh(self, *args, **kwargs):
        return PyBulletAppendCollisionMesh(self.client)(*args, **kwargs)

    @forward_docstring(PyBulletRemoveCollisionMesh)
    def remove_collision_mesh(self, *args, **kwargs):
        return PyBulletRemoveCollisionMesh(self.client)(*args, **kwargs)

    @forward_docstring(PyBulletRemoveAttachedCollisionMesh)
    def remove_attached_collision_mesh(self, *args, **kwargs):
        return PyBulletRemoveAttachedCollisionMesh(self.client)(*args, **kwargs)

    @forward_docstring(PyBulletForwardKinematics)
    def forward_kinematics(self, *args, **kwargs):
        return PyBulletForwardKinematics(self.client)(*args, **kwargs)

    @forward_docstring(PyBulletInverseKinematics)
    def inverse_kinematics(self, *args, **kwargs):
        return PyBulletInverseKinematics(self.client)(*args, **kwargs)



"""
Internal implementation of the planner backend interface for MoveIt!
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.interfaces.client import forward_docstring
from compas_fab.backends.interfaces.client import PlannerInterface
from compas_fab.backends.ros.backend_features import MoveItResetPlanningScene
from compas_fab.backends.ros.backend_features.move_it_add_attached_collision_mesh import MoveItAddAttachedCollisionMesh
from compas_fab.backends.ros.backend_features.move_it_add_collision_mesh import MoveItAddCollisionMesh
from compas_fab.backends.ros.backend_features.move_it_append_collision_mesh import MoveItAppendCollisionMesh
from compas_fab.backends.ros.backend_features.move_it_forward_kinematics import MoveItForwardKinematics
from compas_fab.backends.ros.backend_features.move_it_inverse_kinematics import MoveItInverseKinematics
from compas_fab.backends.ros.backend_features.move_it_plan_cartesian_motion import MoveItPlanCartesianMotion
from compas_fab.backends.ros.backend_features.move_it_plan_motion import MoveItPlanMotion
from compas_fab.backends.ros.backend_features.move_it_planning_scene import MoveItPlanningScene
from compas_fab.backends.ros.backend_features.move_it_remove_attached_collision_mesh import MoveItRemoveAttachedCollisionMesh
from compas_fab.backends.ros.backend_features.move_it_remove_collision_mesh import MoveItRemoveCollisionMesh


class MoveItPlanner(PlannerInterface):
    """Implement the planner backend interface based on MoveIt!
    """

    def __init__(self, client):
        super(MoveItPlanner, self).__init__(client)

    @forward_docstring(MoveItForwardKinematics)
    def forward_kinematics(self, *args, **kwargs):
        return MoveItForwardKinematics(self.client)(*args, **kwargs)

    @forward_docstring(MoveItInverseKinematics)
    def inverse_kinematics(self, *args, **kwargs):
        return MoveItInverseKinematics(self.client)(*args, **kwargs)

    @forward_docstring(MoveItPlanCartesianMotion)
    def plan_cartesian_motion(self, *args, **kwargs):
        return MoveItPlanCartesianMotion(self.client)(*args, **kwargs)

    @forward_docstring(MoveItPlanMotion)
    def plan_motion(self, *args, **kwargs):
        return MoveItPlanMotion(self.client)(*args, **kwargs)

    @forward_docstring(MoveItPlanningScene)
    def get_planning_scene(self, *args, **kwargs):
        return MoveItPlanningScene(self.client)(*args, **kwargs)

    @forward_docstring(MoveItResetPlanningScene)
    def reset_planning_scene(self, *args, **kwargs):
        return MoveItResetPlanningScene(self.client)(*args, **kwargs)

    @forward_docstring(MoveItAddCollisionMesh)
    def add_collision_mesh(self, *args, **kwargs):
        return MoveItAddCollisionMesh(self.client)(*args, **kwargs)

    @forward_docstring(MoveItRemoveCollisionMesh)
    def remove_collision_mesh(self, *args, **kwargs):
        return MoveItRemoveCollisionMesh(self.client)(*args, **kwargs)

    @forward_docstring(MoveItAppendCollisionMesh)
    def append_collision_mesh(self, *args, **kwargs):
        return MoveItAppendCollisionMesh(self.client)(*args, **kwargs)

    @forward_docstring(MoveItAddAttachedCollisionMesh)
    def add_attached_collision_mesh(self, *args, **kwargs):
        return MoveItAddAttachedCollisionMesh(self.client)(*args, **kwargs)

    @forward_docstring(MoveItRemoveAttachedCollisionMesh)
    def remove_attached_collision_mesh(self, *args, **kwargs):
        return MoveItRemoveAttachedCollisionMesh(self.client)(*args, **kwargs)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.exceptions import BackendError

__all__ = [
    'RosError',
    'RosValidationError',
]


class RosError(BackendError):
    """Wraps an exception that occurred on the communication with ROS."""

    def __init__(self, message, error_code):
        super(RosError, self).__init__('Error code: ' +
                                       str(error_code) +
                                       '; ' + message)
        self.error_code = error_code


class RosValidationError(BackendError):
    """Wraps an exception that occurred on validation of a ROS response."""

    def __init__(self, original_exception, response):
        super(RosValidationError, self).__init__(str(original_exception))
        self.response = response
        self.original_exception = original_exception



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.interfaces import AddAttachedCollisionMesh
from compas_fab.backends.vrep.helpers import DEFAULT_OP_MODE
from compas_fab.backends.vrep.helpers import VrepError
from compas_fab.backends.vrep.helpers import assert_robot
from compas_fab.utilities import LazyLoader

__all__ = [
    'VrepAddAttachedCollisionMesh',
]

vrep = LazyLoader('vrep', globals(), 'compas_fab.backends.vrep.remote_api.vrep')


class VrepAddAttachedCollisionMesh(AddAttachedCollisionMesh):
    """Callable to add a building member to the 3D scene and attach it to the robot.
    """
    def __init__(self, client):
        self.client = client

    def add_attached_collision_mesh(self, attached_collision_mesh, options=None):
        """Adds a building member to the 3D scene and attaches it to the robot.

        Args:
            attached_collision_mesh (:class:`compas.datastructures.Mesh`): Mesh
                of the building member that will be attached to the robot.
            options (:obj:`dict`): Dictionary containing the following key-value pairs:

                - ``"robot_name'``: (:obj:`str`) Name of robot instance to attach the building member to.


        Returns:
            int: Object handle (identifier) assigned to the building member.

        .. note::
            All meshes are automatically removed from the scene when the simulation ends.
        """
        options = options or {}
        robot_name = options['robot_name']
        return self.add_building_member(robot_name, attached_collision_mesh)

    def add_building_member(self, robot_name, building_member_mesh):
        """Adds a building member to the 3D scene and attaches it to the robot.

        Args:
            robot_name (:obj:`str`): Name of robot instance to attach the building member to.
            building_member_mesh (:class:`compas.datastructures.Mesh`): Mesh
                of the building member that will be attached to the robot.

        Returns:
            int: Object handle (identifier) assigned to the building member.

        .. note::
            All meshes are automatically removed from the scene when the simulation ends.
        """
        handles = self.client.add_collision_mesh([building_member_mesh])

        if len(handles) != 1:
            raise VrepError('Expected one handle, but multiple found=' + str(handles), -1)

        handle = handles[0]

        parent_handle = self.client.get_object_handle('customGripper' + robot_name + '_connection')
        vrep.simxSetObjectParent(self.client.client_id, handle, parent_handle, True, DEFAULT_OP_MODE)

        return handle

    def pick_building_member(self, robot, building_member_mesh, pickup_frame, metric_values=None):
        """Picks up a building member and attaches it to the robot.

        Args:
            robot (:class:`compas_fab.robots.Robot`): Robot instance to use for pick up.
            building_member_mesh (:class:`compas.datastructures.Mesh`): Mesh
                of the building member that will be attached to the robot.
            pickup_frame (:class:`Frame`): Pickup frame.
            metric_values (:obj:`list` of :obj:`float`): List containing one value
                per configurable joint. Each value ranges from 0 to 1,
                where 1 indicates the axis/joint is blocked and cannot
                move during inverse kinematic solving.

        Returns:
            int: Object handle (identifier) assigned to the building member.
        """
        assert_robot(robot)

        joints = len(robot.get_configurable_joints())
        if not metric_values:
            metric_values = [0.1] * joints

        self.client.set_robot_pose(robot, pickup_frame)

        return self.add_attached_collision_mesh(building_member_mesh, options={'robot_name': robot.name})



"""
*******************************************************************************
compas_fab.backends.ros
*******************************************************************************

.. module:: compas_fab.backends.ros

Package with functionality to interact with `ROS <https://ros.org/>`_.

.. autosummary::
    :toctree: generated/

    RosClient
    RosFileServerLoader
    RosError
    RosValidationError
    MoveItPlanner

"""

from __future__ import absolute_import

from .client import *                     # noqa: F401,F403
from .direct_ur_action_client import *    # noqa: F401,F403
from .exceptions import *                 # noqa: F401,F403
from .fileserver_loader import *          # noqa: F401,F403
from .messages import *                   # noqa: F401,F403
from .planner import *                    # noqa: F401,F403

__all__ = [name for name in dir() if not name.startswith('_')]



import math

__all__ = [
    'map_range',
    'range_geometric_row',
    'arange',
    'diffs',
    'allclose',
    'argsort',
    'argmin',
    'argmax',
    'clamp',
]


def map_range(value, from_min, from_max, to_min, to_max):
    """Performs a linear interpolation of a value within the range of [from_min,
        from_max] to another range of [to_min, to_max].
    """
    from_range = from_max - from_min
    to_range = to_max - to_min
    value_scaled = (value - from_min) / float(from_range)
    return to_min + (value_scaled * to_range)


def range_geometric_row(number, d, r=1.1):
    """Returns a list of numbers with a certain relation to each other.

    The function divides one number into a list of d numbers [n0, n1, ...], such
    that their sum is number and the relation between the numbers is defined
    with n1 = n0 / r, n2 = n1 / r, n3 = n2 / r, ...
    """
    if r <= 0:
        raise ValueError("r must be > 0")

    n0 = number / ((1 - (1 / r)**d) / (1 - 1 / r))

    numbers = [n0]
    for i in range(d - 1):
        numbers.append(numbers[-1] / r)
    return numbers


def arange(start, stop, step):
    """Returns evenly spaced values within a given interval.

    The function is similar to NumPy's *arange* function.
    """
    if math.fabs(stop - (start + step)) > math.fabs(stop - start):
        raise ValueError("Please check the sign of step.")

    len = int(math.ceil((stop - start) / float(step)))
    return [start + i * step for i in range(len)]


def diffs(l1, l2):
    """Returns the element-wise differences between two lists.

    Raises
    ------
    ValueError
        If 2 lists of different length are passed.
    """
    if len(l1) != len(l2):
        raise ValueError("Pass 2 lists of equal length.")
    return [math.fabs(a - b) for a, b in zip(l1, l2)]


def allclose(l1, l2, tol=1e-05):
    """Returns True if two lists are element-wise equal within a tolerance.

    The function is similar to NumPy's *allclose* function.
    """
    for a, b in zip(l1, l2):
        if math.fabs(a - b) > tol:
            return False
    return True


def argsort(numbers):
    """Returns the indices that would sort an array of numbers.

    The function is similar to NumPy's *argsort* function.

    Note
    ----
    For a large list of numbers reconsider using NumPy's *argsort* function,
    since this function might take too long.
    """
    return [i[0] for i in sorted(enumerate(numbers), key=lambda x:x[1])]


def argmin(numbers):
    """Returns the index of the minimum value in numbers.

    The function is similar to NumPy's *argmin* function.

    Note
    ----
    For a large list of numbers reconsider using NumPy's *argmin* function,
    since this function might take too long.
    """
    return argsort(numbers)[0]


def argmax(numbers):
    """Returns the index of the maximum value in numbers.

    The function is similar to NumPy's *argmax* function.

    Note
    ----
    For a large list of numbers reconsider using NumPy's *argmax* function,
    since this function might take too long.
    """
    return argsort(numbers)[-1]


def clamp(value, min_value, max_value):
    """Clamps a value witin the bound [min_value, max_value]

    Returns
    -------
    float
    """
    if min_value > max_value:
        raise ValueError("min_value must be bigger than max_value")
    return float(min(max(value, min_value), max_value))


if __name__ == "__main__":
    print(map_range(2, 0, 10, 0, 100))
    print(arange(3, -4, -0.2))
    print(argsort([34, 1, 7, 2, 100]))
    print(clamp(5, 1, 4))
    print(clamp(0, 1, 4))
    print(clamp(3, 1, 4))



"""
Create a planning scene.

COMPAS FAB v0.18.2
"""
from ghpythonlib.componentbase import executingcomponent as component
from scriptcontext import sticky as st

from compas_fab.ghpython.components import create_id
from compas_fab.robots import PlanningScene


class PlanningSceneComponent(component):
    def RunScript(self, robot):
        key = create_id(self, 'planning_scene')
        if robot:
            st[key] = PlanningScene(robot)
        return st.get(key, None)



"""
Interface definition for planner backend implementations.

This is only internal, these interfaces and their implementations
are managed internally by the RosClient class.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import types

import compas
from roslibpy import Service
from roslibpy import ServiceRequest

from compas_fab.backends.ros.exceptions import RosValidationError

__all__ = [
    'ServiceDescription'
]


class ServiceDescription(object):
    """Internal class to simplify service call code."""

    def __init__(self, name, service_type, request_class=None, response_class=None, validator=None):
        self.name = name
        self.type = service_type
        self.request_class = request_class
        self.response_class = response_class
        self.validator = validator

    def call(self, client, request, callback, errback):
        def inner_handler(response_msg):
            if not compas.PY3:
                # IronPython sometimes decides it's a good idea to use an old-style class
                # to pass the response_msg around, so we need to make sure we turn it into
                # a proper ServiceResponse again
                # https://github.com/compas-dev/compas_fab/issues/235
                is_old_style_class = isinstance(response_msg, types.InstanceType)
                if is_old_style_class:
                    response_msg = dict(response_msg)
            response_object = self.response_class.from_msg(response_msg)

            # Validate the response if there's a validator function assigned
            if self.validator:
                try:
                    self.validator(response_object)
                except Exception as e:
                    errback(RosValidationError(e, response_object))
                    return
            callback(response_object)

        if isinstance(request, tuple):
            request_msg = self.request_class(*request)
        else:
            request_msg = self.request_class(**request)
        srv = Service(client, self.name, self.type)
        srv.call(ServiceRequest(request_msg.msg),
                 callback=inner_handler, errback=errback)

    def __call__(self, client, request, callback, errback):
        return self.call(client, request, callback, errback)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.ros.exceptions import RosError
from compas_fab.backends.ros.messages import Constraints
from compas_fab.backends.ros.messages import JointConstraint
from compas_fab.backends.ros.messages import MoveItErrorCodes
from compas_fab.backends.ros.messages import PositionConstraint
from compas_fab.backends.ros.messages import OrientationConstraint
from compas_fab.robots import Duration
from compas_fab.robots import JointTrajectoryPoint


def validate_response(response):
    """Raise an exception if the response indicates an error condition."""
    if response.error_code != MoveItErrorCodes.SUCCESS:
        raise RosError(response.error_code.human_readable,
                       int(response.error_code))


def convert_constraints_to_rosmsg(constraints, header):
    """Convert COMPAS FAB constraints into ROS Messages."""
    if not constraints:
        return None

    ros_constraints = Constraints()
    for c in constraints:
        if c.type == c.JOINT:
            ros_constraints.joint_constraints.append(
                JointConstraint.from_joint_constraint(c))
        elif c.type == c.POSITION:
            ros_constraints.position_constraints.append(
                PositionConstraint.from_position_constraint(header, c))
        elif c.type == c.ORIENTATION:
            ros_constraints.orientation_constraints.append(
                OrientationConstraint.from_orientation_constraint(header, c))
        else:
            raise NotImplementedError

    return ros_constraints


def convert_trajectory_points(points, joint_types):
    result = []

    for pt in points:
        jtp = JointTrajectoryPoint(joint_values=pt.positions,
                                   joint_types=joint_types,
                                   velocities=pt.velocities,
                                   accelerations=pt.accelerations,
                                   effort=pt.effort,
                                   time_from_start=Duration(pt.time_from_start.secs, pt.time_from_start.nsecs))

        result.append(jtp)

    return result



from __future__ import absolute_import


class URmsg(object):
    """
    """

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __str__(self):
        result = ""
        for key, value in self.__dict__.items():
            if value:
                result += "%s, " % value
        return result[:-2]

    def __repr__(self):
        return self.__str__


class Point(URmsg):
    """
    """

    def __init__(self, x, y, z):
        self.x = x  # [m]
        self.y = y  # [m]
        self.z = z  # [m]

    def __str__(self):
        return '%.6f, %.6f, %.6f' % (self.x, self.y, self.z)


class AxisAngle(URmsg):
    """
    """

    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def __str__(self):
        return '%.6f, %.6f, %.6f' % (self.x, self.y, self.z)


class URPose(URmsg):
    """
    """

    def __init__(self, position=Point(0, 0, 0), orientation=AxisAngle(0, 0, 0)):
        self.position = position
        self.orientation = orientation

    @classmethod
    def from_frame(cls, frame):
        point = frame.point
        ax, ay, az = frame.axis_angle_vector
        return cls(Point(*list(point)), AxisAngle(ax, ay, az))

    def __str__(self):
        return "p[%s, %s]" % (self.position, self.orientation)


class URPoseTrajectoryPoint(URmsg):
    """
    """

    def __init__(self, pose=URPose(), acceleration=None, velocity=None, time=None, radius=None):
        self.pose = pose
        self.acceleration = acceleration  # [m/s^2]
        self.velocity = velocity  # [m/s]
        self.time = time  # [s]
        self.radius = radius  # [m]

    def __str__(self):
        result = "%s" % self.pose
        if self.acceleration:
            result += ", a=%.6f" % self.acceleration
        if self.velocity:
            result += ", v=%.6f" % self.velocity
        if self.time:
            result += ", t=%.6f" % self.time
        if self.radius:
            result += ", r=%.6f" % self.radius
        return result


class URMovej(URmsg):
    """
    """

    def __init__(self, pose_trajectory_point=URPoseTrajectoryPoint()):
        self.pose_trajectory_point = pose_trajectory_point

    def __str__(self):
        return "movej(%s)" % self.pose_trajectory_point


class URMovel(URmsg):
    """
    """

    def __init__(self, pose_trajectory_point=URPoseTrajectoryPoint()):
        self.pose_trajectory_point = pose_trajectory_point

    def __str__(self):
        return "movel(%s)" % self.pose_trajectory_point


class URGoal(URmsg):
    """
    """

    def __init__(self, script_lines=[]):
        self.script = "def prog():\n\t"
        self.script += "\n\t".join([str(line) for line in script_lines])
        self.script += "\nend\nprog()\n\n"

    @property
    def msg(self):
        return {"script": self.script}


if __name__ == "__main__":

    from compas.geometry import Frame
    f1 = Frame((0., -193, 1001.), (-1., 0., 0.), (0., 0., -1.))
    f2 = Frame((0., -193, 709.0), (-1., 0., 0.), (0., 0., -1.))
    f1.point /= 1000.
    f2.point /= 1000.
    frames = [f1, f2]

    acceleration = 0.35
    velocity = 0.17
    time = 5.
    script_lines = []

    for frame in frames:
        ptp = URPoseTrajectoryPoint(URPose.from_frame(
            frame), acceleration, velocity, time, None)
        move = URMovej(ptp)
        script_lines.append(move)

    urgoal = URGoal(script_lines)
    print(urgoal.msg["script_lines"])



"""
Connect or disconnect to ROS.

COMPAS FAB v0.18.2
"""
from ghpythonlib.componentbase import executingcomponent as component
from scriptcontext import sticky as st

from compas_fab.backends import RosClient
from compas_fab.ghpython.components import create_id


class ROSConnect(component):
    def RunScript(self, ip, port, connect):
        ros_client = None

        ip = ip or '127.0.0.1'
        port = port or 9090

        key = create_id(self, 'ros_client')
        ros_client = st.get(key, None)

        if ros_client:
            st[key].close()
        if connect:
            st[key] = RosClient(ip, port)
            st[key].run(5)

        ros_client = st.get(key, None)
        is_connected = ros_client.is_connected if ros_client else False
        return (ros_client, is_connected)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__all__ = [
    'BackendError',
]


class BackendError(Exception):
    """Indicates an exceptional state that caused an error
    within the backend engine."""

    def __init__(self, message):
        super(BackendError, self).__init__(message)



"""
Add an attached collision mesh to the robot.

COMPAS FAB v0.18.2
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_rhino.geometry import RhinoMesh
from compas_fab.robots import AttachedCollisionMesh
from compas_fab.robots import CollisionMesh


class AttachedCollisionMeshComponent(component):
    def RunScript(self, scene, mesh, identifier, link_name, touch_links, add, remove):
        attached_collision_mesh = None
        if scene and mesh and identifier and link_name:
            compas_mesh = RhinoMesh.from_geometry(mesh).to_compas()
            collision_mesh = CollisionMesh(compas_mesh, identifier)
            attached_collision_mesh = AttachedCollisionMesh(collision_mesh, link_name, touch_links)
            if add:
                scene.add_attached_collision_mesh(attached_collision_mesh)
            if remove:
                scene.remove_attached_collision_mesh(identifier)
                scene.remove_collision_mesh(identifier)
        return attached_collision_mesh



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.interfaces import ForwardKinematics


class PyBulletForwardKinematics(ForwardKinematics):
    """Callable to calculate the robot's forward kinematic."""
    def __init__(self, client):
        self.client = client

    def forward_kinematics(self, robot, configuration, group=None, options=None):
        """Calculate the robot's forward kinematic.

        Parameters
        ----------
        robot : :class:`compas_fab.robots.Robot`
            The robot instance for which inverse kinematics is being calculated.
        configuration : :class:`compas_fab.robots.Configuration`
            The full configuration to calculate the forward kinematic for. If no
            full configuration is passed, the zero-joint state for the other
            configurable joints is assumed.
        group : str, optional
            The planning group used for determining the end effector and labeling
            the ``configuration``. Defaults to the robot's main planning group.
        options : dict, optional
            Dictionary containing the following key-value pairs:

            - ``"link"``: (:obj:`str`, optional) The name of the link to
              calculate the forward kinematics for. Defaults to the end effector.
            - ``"check_collision"``: (:obj:`str`, optional) When ``True``,
              :meth:`compas_fab.backends.PyBulletClient.check_collisions` will be called.
              Defaults to ``False``.

        Returns
        -------
        :class:`Frame`
            The frame in the world's coordinate system (WCF).
        """
        link_name = options.get('link') or robot.get_end_effector_link_name(group)
        cached_robot = self.client.get_cached_robot(robot)
        body_id = self.client.get_uid(cached_robot)
        link_id = self.client._get_link_id_by_name(link_name, cached_robot)
        self.client.set_robot_configuration(robot, configuration, group)
        frame = self.client._get_link_frame(link_id, body_id)
        if options.get('check_collision'):
            self.client.collision_check()
        return frame



from __future__ import absolute_import

from compas_fab.backends.ros.messages.geometry_msgs import Point
from compas_fab.backends.ros.messages.geometry_msgs import Pose
from compas_fab.backends.ros.messages.geometry_msgs import PoseStamped
from compas_fab.backends.ros.messages.geometry_msgs import Quaternion
from compas_fab.backends.ros.messages.geometry_msgs import Vector3
from compas_fab.backends.ros.messages.object_recognition_msgs import ObjectType
from compas_fab.backends.ros.messages.octomap_msgs import OctomapWithPose
from compas_fab.backends.ros.messages.sensor_msgs import JointState
from compas_fab.backends.ros.messages.sensor_msgs import MultiDOFJointState
from compas_fab.backends.ros.messages.shape_msgs import Mesh
from compas_fab.backends.ros.messages.shape_msgs import MeshTriangle
from compas_fab.backends.ros.messages.shape_msgs import Plane
from compas_fab.backends.ros.messages.shape_msgs import SolidPrimitive
from compas_fab.backends.ros.messages.std_msgs import Header
from compas_fab.backends.ros.messages.std_msgs import ROSmsg
from compas_fab.backends.ros.messages.trajectory_msgs import JointTrajectory
from compas_fab.backends.ros.messages.trajectory_msgs import MultiDOFJointTrajectory
from compas_fab.robots import AttachedCollisionMesh
from compas_fab.robots import CollisionMesh


class CollisionObject(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/CollisionObject.html
    """
    ADD = 0
    REMOVE = 1
    APPEND = 2
    MOVE = 3

    def __init__(self, header=None, id="collision_obj", type=None,
                 primitives=None, primitive_poses=None, meshes=None, mesh_poses=None,
                 planes=None, plane_poses=None,
                 subframe_names=None, subframe_poses=None, operation=0):
        self.header = header or Header()  # a header, used for interpreting the poses
        self.id = id  # the id of the object (name used in MoveIt)
        self.type = type or ObjectType()  # The object type in a database of known objects
        # solid geometric primitives
        self.primitives = primitives or []
        self.primitive_poses = primitive_poses or []
        # meshes
        self.meshes = meshes or []
        self.mesh_poses = mesh_poses or []
        # bounding planes
        self.planes = planes or []
        self.plane_poses = plane_poses or []

        self.operation = operation  # ADD or REMOVE or APPEND or MOVE

    @classmethod
    def from_collision_mesh(cls, collision_mesh):
        """Creates a collision object from a :class:`compas_fab.robots.CollisionMesh`
        """
        kwargs = {}
        kwargs['header'] = Header(frame_id=collision_mesh.root_name)
        kwargs['id'] = collision_mesh.id
        kwargs['meshes'] = [Mesh.from_mesh(collision_mesh.mesh)]
        kwargs['mesh_poses'] = [Pose.from_frame(collision_mesh.frame)]

        return cls(**kwargs)

    @classmethod
    def from_msg(cls, msg):
        kwargs = {}

        kwargs['header'] = Header.from_msg(msg['header'])
        kwargs['id'] = msg['id']
        kwargs['type'] = ObjectType.from_msg(msg['type'])

        kwargs['primitives'] = [SolidPrimitive.from_msg(i) for i in msg['primitives']]
        kwargs['primitive_poses'] = [Pose.from_msg(i) for i in msg['primitive_poses']]
        kwargs['meshes'] = [Mesh.from_msg(i) for i in msg['meshes']]
        kwargs['mesh_poses'] = [Pose.from_msg(i) for i in msg['mesh_poses']]
        kwargs['planes'] = [Plane.from_msg(i) for i in msg['planes']]
        kwargs['plane_poses'] = [Pose.from_frame(i) for i in msg['plane_poses']]

        kwargs['operation'] = msg['operation']

        return cls(**kwargs)

    def to_collision_meshes(self):
        """Creates a list of collision meshes from a :class:`compas_fab.backends.CollisionObject`
        """
        collision_meshes = []
        for mesh, pose in zip(self.meshes, self.mesh_poses):
            pose = pose if isinstance(pose, Pose) else Pose(**pose)
            pose.position = pose.position if isinstance(pose.position, Point) else Point(**pose.position)
            pose.position.x = float(pose.position.x)
            pose.position.y = float(pose.position.y)
            pose.position.z = float(pose.position.z)
            pose.orientation = pose.orientation if isinstance(pose.orientation, Quaternion) else Quaternion(**pose.orientation)
            pose.orientation.x = float(pose.orientation.x)
            pose.orientation.y = float(pose.orientation.y)
            pose.orientation.z = float(pose.orientation.z)
            pose.orientation.w = float(pose.orientation.w)
            mesh = mesh if isinstance(mesh, Mesh) else Mesh(**mesh)
            mesh.triangles = [t if isinstance(t, MeshTriangle) else MeshTriangle(**t) for t in mesh.triangles]
            for triangle in mesh.triangles:
                triangle.vertex_indices = [int(x) for x in triangle.vertex_indices]
            mesh.vertices = [v if isinstance(v, Point) else Point(**v) for v in mesh.vertices]
            for vertex in mesh.vertices:
                vertex.x = float(vertex.x)
                vertex.y = float(vertex.y)
                vertex.z = float(vertex.z)
            root_name = getattr(self.header, 'frame_id', None) or self.header['frame_id']
            cm = CollisionMesh(mesh.mesh, self.id, pose.frame, root_name)
            collision_meshes.append(cm)
        return collision_meshes


class AttachedCollisionObject(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/AttachedCollisionObject.html
    """

    def __init__(self, link_name=None, object=None, touch_links=None,
                 detach_posture=None, weight=0):
        self.link_name = link_name or ''
        self.object = object or CollisionObject()
        self.touch_links = touch_links or []
        self.detach_posture = detach_posture or JointTrajectory()
        self.weight = weight

    @classmethod
    def from_attached_collision_mesh(cls, attached_collision_mesh):
        """Creates an attached collision object from a :class:`compas_fab.robots.AttachedCollisionMesh`
        """

        kwargs = {}
        kwargs['link_name'] = attached_collision_mesh.link_name
        kwargs['object'] = CollisionObject.from_collision_mesh(attached_collision_mesh.collision_mesh)
        kwargs['touch_links'] = [str(s) for s in attached_collision_mesh.touch_links]
        kwargs['weight'] = attached_collision_mesh.weight

        return cls(**kwargs)

    def to_attached_collision_meshes(self):
        """Creates a list of attached collision meshes from a :class:`compas_fab.backends.AttachedCollisionObject`
        """
        attached_collision_meshes = []
        obj = self.object if isinstance(self.object, CollisionObject) else CollisionObject(**self.object)
        collision_meshes = obj.to_collision_meshes()
        for cm in collision_meshes:
            acm = AttachedCollisionMesh(cm, self.link_name, self.touch_links, self.weight)
            attached_collision_meshes.append(acm)
        return attached_collision_meshes


class Constraints(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/Constraints.html
    """

    def __init__(self, name='', joint_constraints=None, position_constraints=None,
                 orientation_constraints=None, visibility_constraints=None):
        self.name = name
        self.joint_constraints = joint_constraints if joint_constraints else []
        self.position_constraints = position_constraints if position_constraints else []
        self.orientation_constraints = orientation_constraints if orientation_constraints else []
        self.visibility_constraints = visibility_constraints if visibility_constraints else []


class RobotState(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/RobotState.html
    """

    def __init__(self, joint_state=None, multi_dof_joint_state=None,
                 attached_collision_objects=None, is_diff=False):
        self.joint_state = joint_state if joint_state else JointState()
        self.multi_dof_joint_state = multi_dof_joint_state if multi_dof_joint_state else MultiDOFJointState()
        self.attached_collision_objects = attached_collision_objects if attached_collision_objects else []
        self.is_diff = is_diff

    @classmethod
    def from_msg(cls, msg):
        joint_state = JointState.from_msg(msg['joint_state'])
        multi_dof_joint_state = MultiDOFJointState.from_msg(
            msg['multi_dof_joint_state'])
        attached_collision_objects = [AttachedCollisionObject.from_msg(
            item) for item in msg['attached_collision_objects']]
        return cls(joint_state, multi_dof_joint_state, attached_collision_objects, msg['is_diff'])


class PositionIKRequest(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/PositionIKRequest.html
    """

    def __init__(self, group_name="robot", robot_state=None, constraints=None,
                 pose_stamped=None, timeout=None, attempts=8,
                 avoid_collisions=True):
        self.group_name = group_name
        self.robot_state = robot_state if robot_state else RobotState()
        self.constraints = constraints if constraints else Constraints()
        self.avoid_collisions = avoid_collisions
        self.pose_stamped = pose_stamped if pose_stamped else PoseStamped()
        self.timeout = timeout
        self.attempts = attempts


class RobotTrajectory(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/RobotTrajectory.html
    """

    def __init__(self, joint_trajectory=JointTrajectory(),
                 multi_dof_joint_trajectory=MultiDOFJointTrajectory()):
        self.joint_trajectory = joint_trajectory
        self.multi_dof_joint_trajectory = multi_dof_joint_trajectory

    @classmethod
    def from_msg(cls, msg):
        joint_trajectory = JointTrajectory.from_msg(msg['joint_trajectory'])
        multi_dof_joint_trajectory = MultiDOFJointTrajectory.from_msg(
            msg['multi_dof_joint_trajectory'])
        return cls(joint_trajectory, multi_dof_joint_trajectory)


class MoveItErrorCodes(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/MoveItErrorCodes.html
    """
    # overall behavior
    SUCCESS = 1
    FAILURE = 99999

    PLANNING_FAILED = -1
    INVALID_MOTION_PLAN = -2
    MOTION_PLAN_INVALIDATED_BY_ENVIRONMENT_CHANGE = -3
    CONTROL_FAILED = -4
    UNABLE_TO_AQUIRE_SENSOR_DATA = -5
    TIMED_OUT = -6
    PREEMPTED = -7

    # planning & kinematics request errors
    START_STATE_IN_COLLISION = -10
    START_STATE_VIOLATES_PATH_CONSTRAINTS = -11

    GOAL_IN_COLLISION = -12
    GOAL_VIOLATES_PATH_CONSTRAINTS = -13
    GOAL_CONSTRAINTS_VIOLATED = -14

    INVALID_GROUP_NAME = -15
    INVALID_GOAL_CONSTRAINTS = -16
    INVALID_ROBOT_STATE = -17
    INVALID_LINK_NAME = -18
    INVALID_OBJECT_NAME = -19

    # system errors
    FRAME_TRANSFORM_FAILURE = -21
    COLLISION_CHECKING_UNAVAILABLE = -22
    ROBOT_STATE_STALE = -23
    SENSOR_INFO_STALE = -24

    # kinematics errors
    NO_IK_SOLUTION = -31

    def __init__(self, val=-31):
        self.val = val

    def __int__(self):
        return self.val

    def __eq__(self, other):
        return self.val == other

    def __ne__(self, other):
        return self.val != other

    @property
    def human_readable(self):
        cls = type(self)
        for k, v in cls.__dict__.items():
            if v == self.val:
                return k
        return ''


class PlannerParams(ROSmsg):
    """https://docs.ros.org/melodic/api/moveit_msgs/html/msg/PlannerParams.html
    """

    def __init__(self, keys=None, values=None, descriptions=None):
        self.keys = keys or []                  # parameter names (same size as values)
        self.values = values or []              # parameter values (same size as keys)
        self.descriptions = descriptions or []  # parameter description (can be empty)


class WorkspaceParameters(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/WorkspaceParameters.html
    """

    def __init__(self, header=None, min_corner=None, max_corner=None):
        self.header = header or Header()
        self.min_corner = min_corner or Vector3(-1000, -1000, -1000)
        self.max_corner = max_corner or Vector3(1000, 1000, 1000)


class TrajectoryConstraints(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/TrajectoryConstraints.html
    """

    def __init__(self, constraints=None):
        self.constraints = constraints or []  # Constraints[]


class JointConstraint(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/JointConstraint.html
    """

    def __init__(self, joint_name="", position=0, tolerance_above=0, tolerance_below=0, weight=1.):
        self.joint_name = joint_name
        self.position = float(position)
        self.tolerance_above = float(tolerance_above)
        self.tolerance_below = float(tolerance_below)
        self.weight = float(weight)

    @classmethod
    def from_joint_constraint(cls, joint_constraint):
        """Creates a `JointConstraint` from a :class:`compas_fab.robots.JointConstraint`.
        """
        c = joint_constraint
        return cls(c.joint_name, c.value, c.tolerance_above, c.tolerance_below, c.weight)


class VisibilityConstraint(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/VisibilityConstraint.html
    """

    def __init__(self):
        raise NotImplementedError


class BoundingVolume(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/BoundingVolume.html
    """

    def __init__(self, primitives=None, primitive_poses=None, meshes=None,
                 mesh_poses=None):
        self.primitives = primitives or []            # shape_msgs/SolidPrimitive[]
        self.primitive_poses = primitive_poses or []  # geometry_msgs/Pose[]
        self.meshes = meshes or []                    # shape_msgs/Mesh[]
        self.mesh_poses = mesh_poses or []            # geometry_msgs/Pose[]

    @classmethod
    def from_box(cls, box):
        """Creates a `BoundingVolume` from a :class:`compas.geometry.Box`.

        Parameters
        ----------
        box: `compas.geometry.Box`
        """
        primitive = SolidPrimitive.from_box(box)
        pose = Pose.from_frame(box.frame)
        return cls(primitives=[primitive], primitive_poses=[pose])

    @classmethod
    def from_sphere(cls, sphere):
        """Creates a `BoundingVolume` from a :class:`compas.geometry.Sphere`.

        Parameters
        ----------
        sphere: `compas.geometry.Sphere`
        """
        primitive = SolidPrimitive.from_sphere(sphere)
        pose = Pose(Point(*sphere.point), Quaternion(0, 0, 0, 1))
        return cls(primitives=[primitive], primitive_poses=[pose])

    @classmethod
    def from_mesh(cls, mesh):
        """Creates a `BoundingVolume` from a :class:`compas.datastructures.Mesh`.

        Parameters
        ----------
        sphere: `compas.datastructures.Mesh`
        """
        mesh = Mesh.from_mesh(mesh)
        pose = Pose()
        return cls(meshes=[mesh], mesh_poses=[pose])

    @classmethod
    def from_bounding_volume(cls, bounding_volume):
        """Creates a `BoundingVolume` from a :class:`compas_fab.robots.BoundingVolume`.

        Parameters
        ----------
        bounding_volume: `compas_fab.robots.BoundingVolume`
        """
        if bounding_volume.type == bounding_volume.BOX:
            return cls.from_box(bounding_volume.volume)
        elif bounding_volume.type == bounding_volume.SPHERE:
            return cls.from_sphere(bounding_volume.volume)
        elif bounding_volume.type == bounding_volume.MESH:
            return cls.from_mesh(bounding_volume.volume)
        else:
            raise NotImplementedError


class PositionConstraint(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/PositionConstraint.html
    """

    def __init__(self, header=None, link_name=None, target_point_offset=None,
                 constraint_region=None, weight=None):
        self.header = header or Header()
        self.link_name = link_name or ""
        self.target_point_offset = target_point_offset or Vector3(0., 0., 0.)  # geometry_msgs/Vector3
        self.constraint_region = constraint_region or BoundingVolume()         # moveit_msgs/BoundingVolume
        self.weight = float(weight) or 1.

    @classmethod
    def from_position_constraint(cls, header, position_constraint):
        """Creates a `PositionConstraint` from a :class:`compas_fab.robots.PositionConstraint`.
        """
        constraint_region = BoundingVolume.from_bounding_volume(position_constraint.bounding_volume)
        return cls(header, position_constraint.link_name, None, constraint_region, position_constraint.weight)


class OrientationConstraint(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/OrientationConstraint.html
    """

    def __init__(self, header=None, orientation=None, link_name=None,
                 absolute_x_axis_tolerance=0.0, absolute_y_axis_tolerance=0.0,
                 absolute_z_axis_tolerance=0.0, weight=1):
        """
        Notes
        -----
        The naming of the absolute_x/y/z_axis_tolerances might be misleading:
        If you specify the absolute_x/y/z_axis_tolerances with [0.01, 0.01, 6.3],
        it means that the frame's x-axis and y-axis are allowed to rotate about
        the z-axis by an angle of 6.3 radians, whereas the z-axis can only change
        by 0.01.
        """
        self.header = header or Header()
        self.orientation = orientation or Quaternion()  # geometry_msgs/Quaternion
        self.link_name = link_name or ""
        self.absolute_x_axis_tolerance = float(absolute_x_axis_tolerance)
        self.absolute_y_axis_tolerance = float(absolute_y_axis_tolerance)
        self.absolute_z_axis_tolerance = float(absolute_z_axis_tolerance)
        self.weight = float(weight)

    @classmethod
    def from_orientation_constraint(cls, header, orientation_constraint):
        """Creates a ``OrientationConstraint`` from a :class:`compas_fab.robots.OrientationConstraint`.
        """
        qw, qx, qy, qz = orientation_constraint.quaternion
        ax, ay, az = orientation_constraint.tolerances

        kwargs = {}
        kwargs['header'] = header
        kwargs['orientation'] = Quaternion(qx, qy, qz, qw)
        kwargs['link_name'] = orientation_constraint.link_name
        kwargs['absolute_x_axis_tolerance'] = ax
        kwargs['absolute_y_axis_tolerance'] = ay
        kwargs['absolute_z_axis_tolerance'] = az
        kwargs['weight'] = orientation_constraint.weight

        return cls(**kwargs)


class PlanningSceneComponents(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/msg/PlanningSceneComponents.html
    """
    SCENE_SETTINGS = 1
    ROBOT_STATE = 2
    ROBOT_STATE_ATTACHED_OBJECTS = 4
    WORLD_OBJECT_NAMES = 8
    WORLD_OBJECT_GEOMETRY = 16
    OCTOMAP = 32
    TRANSFORMS = 64
    ALLOWED_COLLISION_MATRIX = 128
    LINK_PADDING_AND_SCALING = 256
    OBJECT_COLORS = 512

    def __init__(self, components=None):
        self.components = components or self.SCENE_SETTINGS

    def __eq__(self, other):
        return self.components == other

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def human_readable(self):
        cls = type(self)
        for k, v in cls.__dict__.items():
            if v == self.components:
                return k
        return ''


class AllowedCollisionMatrix(ROSmsg):
    """https://docs.ros.org/melodic/api/moveit_msgs/html/msg/AllowedCollisionMatrix.html
    """

    def __init__(self, entry_names=None, entry_values=None, default_entry_names=None, default_entry_values=None):
        self.entry_names = entry_names or []  # string[]
        self.entry_values = entry_values or []  # moveit_msgs/AllowedCollisionEntry[]
        self.default_entry_names = default_entry_names or []  # string[]
        self.default_entry_values = default_entry_values or []  # bool[]


class PlanningSceneWorld(ROSmsg):
    """https://docs.ros.org/melodic/api/moveit_msgs/html/msg/PlanningSceneWorld.html
    """

    def __init__(self, collision_objects=None, octomap=None):
        self.collision_objects = collision_objects or []  # collision objects # CollisionObject[]
        self.octomap = octomap or OctomapWithPose()  # octomap_msgs/OctomapWithPose

    @classmethod
    def from_msg(cls, msg):
        collision_objects = [CollisionObject.from_msg(i) for i in msg['collision_objects']]
        octomap = msg['octomap']  # TODO: Add OctomapWithPose.from_msg(msg['octomap'])

        return cls(collision_objects, octomap)


class PlanningScene(ROSmsg):
    """https://docs.ros.org/melodic/api/moveit_msgs/html/msg/PlanningScene.html
    """

    def __init__(self, name='', robot_state=None, robot_model_name='',
                 fixed_frame_transforms=None, allowed_collision_matrix=None,
                 link_padding=None, link_scale=None, object_colors=None, world=None,
                 is_diff=False):
        self.name = name                                            # string
        self.robot_state = robot_state or RobotState()              # moveit_msgs/RobotState
        self.robot_model_name = robot_model_name                    # string
        self.fixed_frame_transforms = fixed_frame_transforms or []  # geometry_msgs/TransformStamped[]
        self.allowed_collision_matrix = allowed_collision_matrix or AllowedCollisionMatrix()
        self.link_padding = link_padding or []                      # moveit_msgs/LinkPadding[]
        self.link_scale = link_scale or []                          # moveit_msgs/LinkScale[]
        self.object_colors = object_colors or []                    # moveit_msgs/ObjectColor[]
        self.world = world or PlanningSceneWorld()                  # moveit_msgs/PlanningSceneWorld
        self.is_diff = is_diff                                      # bool

    @classmethod
    def from_msg(cls, msg):
        robot_state = RobotState.from_msg(msg['robot_state'])
        allowed_collision_matrix = msg['allowed_collision_matrix']
        world = PlanningSceneWorld.from_msg(msg['world'])

        return cls(msg['name'], robot_state, msg['robot_model_name'],
                   msg['fixed_frame_transforms'], allowed_collision_matrix,
                   msg['link_padding'], msg['link_scale'], msg['object_colors'],
                   world, msg['is_diff'])


class ExecuteTrajectoryGoal(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/action/ExecuteTrajectory.html
    """

    def __init__(self, trajectory=None):
        self.trajectory = trajectory or RobotTrajectory()


class ExecuteTrajectoryFeedback(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/action/ExecuteTrajectory.html
    """

    def __init__(self, state=None):
        self.state = state

    @classmethod
    def from_msg(cls, msg):
        return cls(msg['state'])


class ExecuteTrajectoryResult(ROSmsg):
    """https://docs.ros.org/kinetic/api/moveit_msgs/html/action/ExecuteTrajectory.html
    """

    def __init__(self, error_code=None):
        self.error_code = error_code or MoveItErrorCodes()  # moveit_msgs/MoveItErrorCodes

    @classmethod
    def from_msg(cls, msg):
        error_code = MoveItErrorCodes.from_msg(msg['error_code'])
        return cls(error_code)



"""Classes used to define robot trajectories."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas.robots import Configuration
from compas.robots.configuration import FixedLengthList
from compas_fab.robots import AttachedCollisionMesh
from compas_fab.robots.time_ import Duration

__all__ = [
    'JointTrajectoryPoint',
    'JointTrajectory',
    'Trajectory',
]


class JointTrajectoryPoint(Configuration):
    """Defines a point within a trajectory.

    A trajectory point is a sub-class of :class:`Configuration` extended
    with acceleration, effort and time from start information.

    Trajectory points are defined either as *joint_values + velocities and
    accelerations*, or as *joint_values + efforts*.

    Parameters
    ----------
    joint_values : :obj:`list` of :obj:`float`, optional
        Joint values expressed in radians or meters, depending on the respective
        type.
    joint_types : :obj:`list` of :attr:`compas.robots.Joint.TYPE`, optional
        Joint types, e.g. a :obj:`list` of
        :attr:`compas.robots.Joint.REVOLUTE` for revolute joints.
    velocities : :obj:`list` of :obj:`float`, optional
        Velocity of each joint.
    accelerations : :obj:`list` of :obj:`float`, optional
        Acceleration of each joint.
    effort : :obj:`list` of :obj:`float`, optional
        Effort of each joint.
    time_from_start : :class:`Duration`, optional
        Duration of trajectory point counting from the start.

    Attributes
    ----------
    joint_values : :obj:`list` of :obj:`float`
        Joint values expressed in radians or meters, depending on the respective
        type.
    joint_types : :obj:`list` of :attr:`compas.robots.Joint.TYPE`
        Joint types, e.g. a :obj:`list` of
        :attr:`compas.robots.Joint.REVOLUTE` for revolute joints.
    velocities : :obj:`list` of :obj:`float`
        Velocity of each joint.
    accelerations : :obj:`list` of :obj:`float`
        Acceleration of each joint.
    effort : :obj:`list` of :obj:`float`
        Effort of each joint.
    time_from_start : :class:`Duration`
        Duration of trajectory point counting from the start.
    positions : :obj:`list` of :obj:`float`
        Alias of `joint_values`.
    data : obj:`dict`
        The data representing the trajectory point.
    """

    def __init__(self, joint_values=None, joint_types=None, velocities=None, accelerations=None, effort=None, time_from_start=None, joint_names=None):
        super(JointTrajectoryPoint, self).__init__(joint_values, joint_types, joint_names)
        self.velocities = velocities or len(self.joint_values) * [0.]
        self.accelerations = accelerations or len(self.joint_values) * [0.]
        self.effort = effort or len(self.joint_values) * [0.]
        self.time_from_start = time_from_start or Duration(0, 0)

    def __str__(self):
        """Return a human-readable string representation of the instance."""
        vs = '%.' + self._precision
        return 'JointTrajectoryPoint(({}), {}, ({}), ({}), ({}), {})'.format(
            ', '.join(vs % i for i in self.joint_values),
            tuple(self.joint_types),
            ', '.join(vs % i for i in self.velocities),
            ', '.join(vs % i for i in self.accelerations),
            ', '.join(vs % i for i in self.effort),
            self.time_from_start,
        )

    @property
    def positions(self):
        """:obj:`list` of :obj:`float` : Alias of `joint_values`."""
        return self.joint_values

    @property
    def velocities(self):
        """:obj:`list` of :obj:`float` : Velocity of each joint."""
        return self._velocities

    @velocities.setter
    def velocities(self, velocities):
        if len(self.joint_values) != len(velocities):
            raise ValueError('Must have {} velocities, but {} given.'.format(
                len(self.joint_values), len(velocities)))

        self._velocities = FixedLengthList(velocities)

    @property
    def accelerations(self):
        """:obj:`list` of :obj:`float` : Acceleration of each joint."""
        return self._accelerations

    @accelerations.setter
    def accelerations(self, accelerations):
        if len(self.joint_values) != len(accelerations):
            raise ValueError('Must have {} accelerations, but {} given.'.format(
                len(self.joint_values), len(accelerations)))

        self._accelerations = FixedLengthList(accelerations)

    @property
    def effort(self):
        """:obj:`list` of :obj:`float` : Effort of each joint."""
        return self._effort

    @effort.setter
    def effort(self, effort):
        if len(self.joint_values) != len(effort):
            raise ValueError('Must have {} efforts, but {} given.'.format(
                len(self.joint_values), len(effort)))

        self._effort = FixedLengthList(effort)

    @property
    def data(self):
        """:obj:`dict` : The data representing the trajectory point.

        By assigning a data dictionary to this property, the current data of the
        configuration will be replaced by the data in the :obj:`dict`. The data getter
        and setter should always be used in combination with each other.
        """
        data_obj = super(JointTrajectoryPoint, self).data
        data_obj['velocities'] = self.velocities
        data_obj['accelerations'] = self.accelerations
        data_obj['effort'] = self.effort
        data_obj['time_from_start'] = self.time_from_start.to_data()

        return data_obj

    @data.setter
    def data(self, data):
        self._joint_values = FixedLengthList(data.get('joint_values') or [])
        self._joint_types = FixedLengthList(data.get('joint_types') or [])
        self._joint_names = FixedLengthList(data.get('joint_names') or [])
        self._velocities = FixedLengthList(data.get('velocities') or [])
        self._accelerations = FixedLengthList(data.get('accelerations') or [])
        self._effort = FixedLengthList(data.get('effort') or [])
        self.time_from_start = Duration.from_data(data.get('time_from_start') or {})

    def copy(self):
        """Create a copy of this :class:`JointTrajectoryPoint`.

        Returns
        -------
        :class:`JointTrajectoryPoint`
            An instance of :class:`JointTrajectoryPoint`
        """
        cls = type(self)
        return cls.from_data(self.data)

    @property
    def velocity_dict(self):
        """A dictionary of joint velocities by joint name."""
        self.check_joint_names()
        return dict(zip(self.joint_names, self.velocities))

    @property
    def acceleration_dict(self):
        """A dictionary of joint accelerations by joint name."""
        self.check_joint_names()
        return dict(zip(self.joint_names, self.accelerations))

    @property
    def effort_dict(self):
        """A dictionary of joint efforts by joint name."""
        self.check_joint_names()
        return dict(zip(self.joint_names, self.effort))

    def merged(self, other):
        """Get a new ``JointTrajectoryPoint`` with this ``JointTrajectoryPoint`` merged
        with another ``JointTrajectoryPoint``.  The other ``JointTrajectoryPoint``
        takes precedence over this ``JointTrajectoryPoint`` in case a joint value is present in both.

        Note
        ----
            Caution: ``joint_names`` may be rearranged.

        Parameters
        ----------
        other : :class:`JointTrajectoryPoint`
            The ``JointTrajectoryPoint`` to be merged.

        Returns
        -------
        :class:`JointTrajectoryPoint`
            A ``JointTrajectoryPoint`` with values for all included joints.

        Raises
        ------
        :exc:`ValueError`
            If the ``JointTrajectoryPoint`` or the other ``JointTrajectoryPoint`` does not specify
            joint names for all joint values.
        """
        _joint_dict = self.joint_dict
        _joint_dict.update(other.joint_dict)

        _type_dict = self.type_dict
        _type_dict.update(other.type_dict)

        _velocity_dict = self.velocity_dict
        _velocity_dict.update(other.velocity_dict)

        _acceleration_dict = self.acceleration_dict
        _acceleration_dict.update(other.acceleration_dict)

        _effort_dict = self.effort_dict
        _effort_dict.update(other.effort_dict)

        joint_names = list(_joint_dict.keys())
        joint_values = [_joint_dict[name] for name in joint_names]
        joint_types = [_type_dict[name] for name in joint_names]
        velocities = [_velocity_dict[name] for name in joint_names]
        accelerations = [_acceleration_dict[name] for name in joint_names]
        effort = [_effort_dict[name] for name in joint_names]

        return JointTrajectoryPoint(joint_values, joint_types, velocities, accelerations, effort, joint_names=joint_names)


class Trajectory(object):
    """Base trajectory class.

    Attribute
    ---------
    planning_time : :obj:`float`
        Amount of time it took to complete the motion plan
    """

    def __init__(self):
        self.planning_time = None


class JointTrajectory(Trajectory):
    """Describes a joint trajectory as a list of trajectory points.

    Parameters
    ----------
    trajectory_points : :obj:`list` of :class:`JointTrajectoryPoint`, optional
        List of points composing the trajectory.
    joint_names : :obj:`list` of :obj:`str`, optional
        List of joint names of the trajectory.
    start_configuration : :class:`Configuration`, optional
        Start configuration for the trajectory.
    fraction : :obj:`float`, optional
        Indicates the percentage of requested trajectory that was calculated,
        e.g. ``1`` means the full trajectory was found.
    attached_collision_meshes : :obj:`list` of :closs:`compas_fab.robots.AttachedCollisionMesh`
        The attached collision meshes included in the calculation of this trajectory.

    Attributes
    ----------
    points : :obj:`list` of :class:`JointTrajectoryPoint`
        List of points composing the trajectory.
    joint_names : :obj:`list` of :obj:`str`
        List of joint names of the trajectory.
    start_configuration : :class:`Configuration`
        Start configuration for the trajectory.
    fraction : :obj:`float`
        Indicates the percentage of requested trajectory that was calculated,
        e.g. ``1`` means the full trajectory was found.
    attached_collision_meshes : :obj:`list` of :class:`compas_fab.robots.AttachedCollisionMesh`
        The attached collision meshes included in the calculation of this trajectory.
    data : :obj:`dict`
        The data representing the trajectory.
    """

    def __init__(self, trajectory_points=None, joint_names=None, start_configuration=None, fraction=None, attached_collision_meshes=None):
        super(Trajectory, self).__init__()
        self.points = trajectory_points or []
        self.joint_names = joint_names or []
        self.start_configuration = start_configuration
        self.fraction = fraction
        self.attached_collision_meshes = attached_collision_meshes or []

    @classmethod
    def from_data(cls, data):
        """Construct a trajectory from its data representation.

        Parameters
        ----------
        data : :obj:`dict`
            The data dictionary.

        Returns
        -------
        :class:`JointTrajectory`
             An instance of :class:`JointTrajectory`.
        """
        trajectory = cls()
        trajectory.data = data
        return trajectory

    def to_data(self):
        """Get the data dictionary that represents the trajectory.

        This can be used to reconstruct the :class:`Trajectory` instance.

        Returns
        -------
        :obj:`dict`
        """
        return self.data

    @property
    def data(self):
        """:obj:`dict` : The data representing the trajectory."""
        data_obj = {}
        data_obj['points'] = [p.to_data() for p in self.points]
        data_obj['joint_names'] = self.joint_names or []
        data_obj['start_configuration'] = self.start_configuration.to_data() if self.start_configuration else None
        data_obj['fraction'] = self.fraction
        data_obj['attached_collision_meshes'] = [acm.to_data() for acm in self.attached_collision_meshes]

        return data_obj

    @data.setter
    def data(self, data):
        self.points = list(map(JointTrajectoryPoint.from_data, data.get('points') or []))
        self.joint_names = data.get('joint_names', [])
        if data.get('start_configuration'):
            self.start_configuration = Configuration.from_data(data.get('start_configuration'))
        self.fraction = data.get('fraction')
        self.attached_collision_meshes = [AttachedCollisionMesh.from_data(acm_data) for acm_data in data.get('attached_collision_meshes')]

    @property
    def time_from_start(self):
        """:obj:`float` : Effectively, time from start for the last point in the trajectory."""
        if not self.points:
            return 0.

        return self.points[-1].time_from_start.seconds



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


class ForwardKinematics(object):
    """Interface for a Planner's forward kinematics feature.  Any implementation of
    ``ForwardKinematics`` must define the method ``forward_kinematics``.  The
    ``__call__`` magic method allows an instance of an implementation of
    ``ForwardKinematics`` to be treated as its ``forward_kinematics`` method.  See
    <https://docs.python.org/3/reference/datamodel.html#object.__call__> and
    <https://en.wikipedia.org/wiki/Function_object#In_Python>.
    """

    def __call__(self, robot, configuration, group=None, options=None):
        return self.forward_kinematics(robot, configuration, group, options)

    def forward_kinematics(self, robot, configuration, group=None, options=None):
        """Calculate the robot's forward kinematic.

        Parameters
        ----------
        robot : :class:`compas_fab.robots.Robot`
            The robot instance for which forward kinematics is being calculated.
        configuration : :class:`compas_fab.robots.Configuration`
            The full configuration to calculate the forward kinematic for. If no
            full configuration is passed, the zero-joint state for the other
            configurable joints is assumed.
        group : str, optional
            The name of the group to be used in the calculation.
        options : dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        :class:`Frame`
            The frame in the world's coordinate system (WCF).
        """
        pass


class InverseKinematics(object):
    """Interface for a Planner's inverse kinematics feature.  Any implementation of
    ``InverseKinematics`` must define the method ``inverse_kinematics``.  The
    ``__call__`` magic method allows an instance of an implementation of
    ``InverseKinematics`` to be treated as its ``inverse_kinematics`` method.  See
    <https://docs.python.org/3/reference/datamodel.html#object.__call__> and
    <https://en.wikipedia.org/wiki/Function_object#In_Python>.
    """

    def __call__(self, robot, frame_WCF, start_configuration=None, group=None, options=None):
        return self.inverse_kinematics(robot, frame_WCF, start_configuration, group, options)

    def inverse_kinematics(self, robot, frame_WCF, start_configuration=None, group=None, options=None):
        """Calculate the robot's inverse kinematic for a given frame.

        Parameters
        ----------
        robot : :class:`compas_fab.robots.Robot`
            The robot instance for which inverse kinematics is being calculated.
        frame_WCF: :class:`compas.geometry.Frame`
            The frame to calculate the inverse for.
        start_configuration: :class:`compas_fab.robots.Configuration`, optional
        group: str, optional
            The planning group used for calculation.
        options: dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        :obj:`tuple` of :obj:`list`
            A tuple of 2 elements containing a list of joint positions and a list of matching joint names.
        """
        pass


class PlanMotion(object):
    """Interface for a Planner's plan motion feature.  Any implementation of
    ``PlanMotion`` must define the method ``plan_motion``.  The
    ``__call__`` magic method allows an instance of an implementation of
    ``PlanMotion`` to be treated as its ``plan_motion`` method.  See
    <https://docs.python.org/3/reference/datamodel.html#object.__call__> and
    <https://en.wikipedia.org/wiki/Function_object#In_Python>.
    """

    def __call__(self, robot, goal_constraints, start_configuration=None, group=None, options=None):
        return self.plan_motion(robot, goal_constraints, start_configuration, group, options)

    def plan_motion(self, robot, goal_constraints, start_configuration=None, group=None, options=None):
        """Calculates a motion path.

        Parameters
        ----------
        robot : :class:`compas_fab.robots.Robot`
            The robot instance for which the motion path is being calculated.
        goal_constraints: list of :class:`compas_fab.robots.Constraint`
            The goal to be achieved, defined in a set of constraints.
            Constraints can be very specific, for example defining value domains
            for each joint, such that the goal configuration is included,
            or defining a volume in space, to which a specific robot link (e.g.
            the end-effector) is required to move to.
        start_configuration: :class:`compas_fab.robots.Configuration`, optional
            The robot's full configuration, i.e. values for all configurable
            joints of the entire robot, at the starting position.
        group: str, optional
            The name of the group to plan for.
        options : dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        :class:`compas_fab.robots.JointTrajectory`
            The calculated trajectory.
        """
        pass


class PlanCartesianMotion(object):
    """Interface for a Planner's plan cartesian motion feature.  Any implementation of
    ``PlanCartesianMotion`` must define the method ``plan_cartesian_motion``.  The
    ``__call__`` magic method allows an instance of an implementation of
    ``PlanCartesianMotion`` to be treated as its ``plan_cartesian_motion`` method.  See
    <https://docs.python.org/3/reference/datamodel.html#object.__call__> and
    <https://en.wikipedia.org/wiki/Function_object#In_Python>.
    """

    def __call__(self, robot, frames_WCF, start_configuration=None, group=None, options=None):
        return self.plan_cartesian_motion(robot, frames_WCF, start_configuration, group, options)

    def plan_cartesian_motion(self, robot, frames_WCF, start_configuration=None, group=None, options=None):
        """Calculates a cartesian motion path (linear in tool space).

        Parameters
        ----------
        robot : :class:`compas_fab.robots.Robot`
            The robot instance for which the cartesian motion path is being calculated.
        frames_WCF: list of :class:`compas.geometry.Frame`
            The frames through which the path is defined.
        start_configuration: :class:`Configuration`, optional
            The robot's full configuration, i.e. values for all configurable
            joints of the entire robot, at the starting position.
        group: str, optional
            The planning group used for calculation.
        options: dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        :class:`compas_fab.robots.JointTrajectory`
            The calculated trajectory.
        """
        pass


class GetPlanningScene(object):
    """Interface for a Planner's get planning scene feature.  Any implementation of
    ``GetPlanningScene`` must define the method ``get_planning_scene``.  The
    ``__call__`` magic method allows an instance of an implementation of
    ``GetPlanningScene`` to be treated as its ``get_planning_scene`` method.  See
    <https://docs.python.org/3/reference/datamodel.html#object.__call__> and
    <https://en.wikipedia.org/wiki/Function_object#In_Python>.
    """

    def __call__(self, options=None):
        return self.get_planning_scene(options)

    def get_planning_scene(self, options=None):
        """Retrieve the planning scene.

        Parameters
        ----------
        options : dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        :class:`compas_fab.robots.planning_scene.PlanningScene`
        """
        pass


class ResetPlanningScene(object):
    """Interface for a Planner's reset planning scene feature.  Any implementation of
    ``ResetPlanningScene`` must define the method ``reset_planning_scene``.  The
    ``__call__`` magic method allows an instance of an implementation of
    ``ResetPlanningScene`` to be treated as its ``reset_planning_scene`` method.  See
    <https://docs.python.org/3/reference/datamodel.html#object.__call__> and
    <https://en.wikipedia.org/wiki/Function_object#In_Python>.
    """

    def __call__(self, options=None):
        return self.reset_planning_scene(options)

    def reset_planning_scene(self, options=None):
        """Retrieve the planning scene.

        Parameters
        ----------
        options : dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        ``None``
        """
        pass


class AddCollisionMesh(object):
    """Interface for a Planner's add collision mesh feature.  Any implementation of
    ``AddCollisionMesh`` must define the method ``add_collision_mesh``.  The
    ``__call__`` magic method allows an instance of an implementation of
    ``AddCollisionMesh`` to be treated as its ``add_collision_mesh`` method.  See
    <https://docs.python.org/3/reference/datamodel.html#object.__call__> and
    <https://en.wikipedia.org/wiki/Function_object#In_Python>.
    """

    def __call__(self, collision_mesh, options=None):
        return self.add_collision_mesh(collision_mesh, options)

    def add_collision_mesh(self, collision_mesh, options=None):
        """Add a collision mesh to the planning scene.

        Parameters
        ----------
        collision_mesh : :class:`compas_fab.robots.CollisionMesh`
            Object containing the collision mesh to be added.
        options : dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        ``None``
        """
        pass


class RemoveCollisionMesh(object):
    """Interface for a Planner's remove collision mesh feature.  Any implementation of
    ``RemoveCollisionMesh`` must define the method ``remove_collision_mesh``.  The
    ``__call__`` magic method allows an instance of an implementation of
    ``RemoveCollisionMesh`` to be treated as its ``remove_collision_mesh`` method.  See
    <https://docs.python.org/3/reference/datamodel.html#object.__call__> and
    <https://en.wikipedia.org/wiki/Function_object#In_Python>.
    """

    def __call__(self, id, options=None):
        return self.remove_collision_mesh(id, options)

    def remove_collision_mesh(self, id, options=None):
        """Remove a collision mesh from the planning scene.

        Parameters
        ----------
        id : str
            Name of collision mesh to be removed.
        options : dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        ``None``
        """
        pass


class AppendCollisionMesh(object):
    """Interface for a Planner's append collision mesh feature.  Any implementation of
    ``AppendCollisionMesh`` must define the method ``append_collision_mesh``.  The
    ``__call__`` magic method allows an instance of an implementation of
    ``AppendCollisionMesh`` to be treated as its ``append_collision_mesh`` method.  See
    <https://docs.python.org/3/reference/datamodel.html#object.__call__> and
    <https://en.wikipedia.org/wiki/Function_object#In_Python>.
    """

    def __call__(self, collision_mesh, options=None):
        return self.append_collision_mesh(collision_mesh, options)

    def append_collision_mesh(self, collision_mesh, options=None):
        """Append a collision mesh to the planning scene.

        Parameters
        ----------
        collision_mesh : :class:`compas_fab.robots.CollisionMesh`
            Object containing the collision mesh to be appended.
        options : dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        ``None``
        """
        pass


class AddAttachedCollisionMesh(object):
    """Interface for a Planner's add attached collision mesh feature.  Any implementation of
    ``AddAttachedCollisionMesh`` must define the method ``add_attached_collision_mesh``.  The
    ``__call__`` magic method allows an instance of an implementation of
    ``AddAttachedCollisionMesh`` to be treated as its ``add_attached_collision_mesh`` method.  See
    <https://docs.python.org/3/reference/datamodel.html#object.__call__> and
    <https://en.wikipedia.org/wiki/Function_object#In_Python>.
    """

    def __call__(self, attached_collision_mesh, options=None):
        return self.add_attached_collision_mesh(attached_collision_mesh, options)

    def add_attached_collision_mesh(self, attached_collision_mesh, options=None):
        """Add a collision mesh and attach it to the robot.

        Parameters
        ----------
        attached_collision_mesh : :class:`compas_fab.robots.AttachedCollisionMesh`
            Object containing the collision mesh to be attached.
        options : dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        ``None``
        """
        pass


class RemoveAttachedCollisionMesh(object):
    """Interface for a Planner's remove attached collision mesh feature.  Any implementation of
    ``RemoveAttachedCollisionMesh`` must define the method ``remove_attached_collision_mesh``.  The
    ``__call__`` magic method allows an instance of an implementation of
    ``RemoveAttachedCollisionMesh`` to be treated as its ``remove_attached_collision_mesh`` method.  See
    <https://docs.python.org/3/reference/datamodel.html#object.__call__> and
    <https://en.wikipedia.org/wiki/Function_object#In_Python>.
    """

    def __call__(self, id, options=None):
        return self.remove_attached_collision_mesh(id, options)

    def remove_attached_collision_mesh(self, id, options=None):
        """Remove an attached collision mesh from the robot.

        Parameters
        ----------
        id : str
            Name of collision mesh to be removed.
        options : dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        ``None``
        """
        pass



import socket

from compas.geometry import Frame
from compas.geometry import matrix_from_frame
from compas.robots import Configuration

from compas_fab.backends.exceptions import BackendError


__all__ = [
    'VrepError',
]

DEFAULT_OP_MODE = 0x010000  # defined in vrepConst.simx_opmode_blocking, but redefined here to prevent loading the remoteApi library

# --------------------------------------------------------------------------
# MAPPINGS
# The following mapping functions are only internal to make sure
# all transformations from and to V-REP are consistent
# --------------------------------------------------------------------------


def vrep_pose_to_frame(pose, scale):
    return Frame.from_list(floats_from_vrep(pose, scale))


def frame_to_vrep_pose(frame, scale):
    # COMPAS FAB uses meters, just like V-REP,
    # so in general, scale should always be 1
    pose = matrix_from_frame(frame)
    pose[0][3] = pose[0][3] / scale
    pose[1][3] = pose[1][3] / scale
    pose[2][3] = pose[2][3] / scale
    return pose[0] + pose[1] + pose[2] + pose[3]


def config_from_vrep(list_of_floats, scale):
    # COMPAS FAB uses radians and meters, just like V-REP,
    # so in general, scale should always be 1
    radians = list_of_floats[3:]
    prismatic_values = map(lambda v: v * scale, list_of_floats[0:3])
    return Configuration.from_prismatic_and_revolute_values(prismatic_values, radians)


def config_to_vrep(config, scale):
    # COMPAS FAB uses radians and meters, just like V-REP,
    # so in general, scale should always be 1
    values = list(map(lambda v: v / scale, config.prismatic_values))
    values.extend(config.revolute_values)
    return values


def floats_to_vrep(list_of_floats, scale):
    return [v / scale for v in list_of_floats]


def floats_from_vrep(list_of_floats, scale):
    return [v * scale for v in list_of_floats]


def assert_robot(robot):
    if not robot:
        raise ValueError('No instance of robot found')
    if not robot.model:
        raise ValueError('The robot instance has no model information attached')
    if 'index' not in robot.model.attr:
        raise ValueError('Robot model needs to define an index as part of the model.attr dictionary')

# --------------------------------------------------------------------------
# NETWORKING HELPERS
# A couple of simple networking helpers for host name resolution
# --------------------------------------------------------------------------


def is_ipv4_address(addr):
    try:
        socket.inet_aton(addr)
        return True
    except socket.error:
        return False


def resolve_host(host):
    if is_ipv4_address(host):
        return host
    else:
        return socket.gethostbyname(host)


class VrepError(BackendError):
    """Wraps an exception that occurred inside the simulation engine."""

    def __init__(self, message, error_code):
        super(VrepError, self).__init__('Error code: ' +
                                        str(error_code) +
                                        '; ' + message)
        self.error_code = error_code



"""
********************************************************************************
compas_fab.utilities
********************************************************************************

.. currentmodule:: compas_fab.utilities

Package containing a set of utility functions.

File system functions
=====================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    read_csv_to_dictionary
    write_data_to_json
    read_data_from_json
    list_files_in_directory

Numerical functions
===================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    map_range
    arange
    allclose

Other functions
===============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    sign
    argsort
    LazyLoader

"""

from .file_io import *        # noqa: F401,F403
from .filesystem import *     # noqa: F401,F403
from .lazy_loader import *    # noqa: F401,F403
from .numbers import *        # noqa: F401,F403
from .utilities import *      # noqa: F401,F403

__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.interfaces import RemoveCollisionMesh
from compas_fab.backends.vrep.helpers import DEFAULT_OP_MODE
from compas_fab.utilities import LazyLoader

__all__ = [
    'VrepRemoveCollisionMesh',
]

vrep = LazyLoader('vrep', globals(), 'compas_fab.backends.vrep.remote_api.vrep')


class VrepRemoveCollisionMesh(RemoveCollisionMesh):
    """Callable to remove collision meshes from the 3D scene.
    """
    def __init__(self, client):
        self.client = client

    def remove_collision_mesh(self, id, options=None):
        """Removes objects from the 3D scene.

        Args:
            id (:obj:`list` of :obj:`int`) : Object handles to remove.

        .. note::
            Please note there's no need to clean up objects manually after the simulation
            has completed, as those will be reset automatically anyway. This method is
            only useful if you need to remove objects *during* a simulation.
        """
        self.remove_objects(id)

    def remove_meshes(self, mesh_handles):
        """Removes meshes from the 3D scene.

        This is functionally identical to ``remove_objects``, but it's here for
        symmetry reasons.

        Args:
            mesh_handles (:obj:`list` of :obj:`int`): Object handles to remove.
        """
        self.remove_objects(mesh_handles)

    def remove_objects(self, object_handles):
        """Removes objects from the 3D scene.

        Args:
            object_handles (:obj:`list` of :obj:`int`): Object handles to remove.

        .. note::
            Please note there's no need to clean up objects manually after the simulation
            has completed, as those will be reset automatically anyway. This method is
            only useful if you need to remove objects *during* a simulation.
        """
        for handle in object_handles:
            vrep.simxRemoveObject(self.client.client_id, handle, DEFAULT_OP_MODE)

        self.client._added_handles = filter(lambda x: x not in object_handles, self.client._added_handles)



from __future__ import absolute_import

from compas_fab.backends.ros.backend_features.move_it_add_attached_collision_mesh import MoveItAddAttachedCollisionMesh
from compas_fab.backends.ros.backend_features.move_it_add_collision_mesh import MoveItAddCollisionMesh
from compas_fab.backends.ros.backend_features.move_it_append_collision_mesh import MoveItAppendCollisionMesh
from compas_fab.backends.ros.backend_features.move_it_forward_kinematics import MoveItForwardKinematics
from compas_fab.backends.ros.backend_features.move_it_inverse_kinematics import MoveItInverseKinematics
from compas_fab.backends.ros.backend_features.move_it_plan_cartesian_motion import MoveItPlanCartesianMotion
from compas_fab.backends.ros.backend_features.move_it_plan_motion import MoveItPlanMotion
from compas_fab.backends.ros.backend_features.move_it_planning_scene import MoveItPlanningScene
from compas_fab.backends.ros.backend_features.move_it_remove_attached_collision_mesh import MoveItRemoveAttachedCollisionMesh
from compas_fab.backends.ros.backend_features.move_it_remove_collision_mesh import MoveItRemoveCollisionMesh
from compas_fab.backends.ros.backend_features.move_it_reset_planning_scene import MoveItResetPlanningScene

__all__ = [
    'MoveItAddAttachedCollisionMesh',
    'MoveItAddCollisionMesh',
    'MoveItAppendCollisionMesh',
    'MoveItForwardKinematics',
    'MoveItInverseKinematics',
    'MoveItPlanCartesianMotion',
    'MoveItPlanMotion',
    'MoveItPlanningScene',
    'MoveItRemoveAttachedCollisionMesh',
    'MoveItRemoveCollisionMesh',
    'MoveItResetPlanningScene',
]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from timeit import default_timer as timer

from compas_fab.backends.interfaces import PlanMotion
from compas_fab.backends.vrep.helpers import config_from_vrep
from compas_fab.backends.vrep.helpers import config_to_vrep
from compas_fab.backends.vrep.helpers import floats_to_vrep
from compas_fab.backends.vrep.helpers import frame_to_vrep_pose
from compas_fab.backends.vrep.helpers import VrepError

__all__ = [
    'VrepPlanMotion',
]


class VrepPlanMotion(PlanMotion):
    """Callable to find a path plan to move the selected robot from its current position to the `goal_constraints`.
    """
    SUPPORTED_PLANNERS = ('bitrrt', 'bkpiece1', 'est', 'kpiece1',
                          'lazyprmstar', 'lbkpiece1', 'lbtrrt', 'pdst',
                          'prm', 'prrt', 'rrt', 'rrtconnect', 'rrtstar',
                          'sbl', 'stride', 'trrt')

    def __init__(self, client):
        self.client = client

    def plan_motion(self, robot, goal_constraints, start_configuration=None, group=None, options=None):
        """Finds a path plan to move the selected robot from its current position to the `goal_constraints`.

        Args:
            robot (:class:`compas_fab.robots.Robot`): The robot instance for which the motion plan is being calculated.
            goal_constraints (:class:`Frame`): Target or goal frame.
            start_configuration (:obj:`None`): Unused parameter.
            group (:obj:`int`): Integer referencing the desired robot group.
            options (:obj:`dict`): Dictionary containing the following key-values pairs:

                - ``"metric_values"``: (:obj:`list` of :obj:`float`) List containing one value
                  per configurable joint. Each value ranges from 0 to 1,
                  where 1 indicates the axis/joint is blocked and cannot
                  move during inverse kinematic solving.
                - ``"collision_meshes"``: (:obj:`list` of :class:`compas.datastructures.Mesh`) Collision meshes
                  to be taken into account when calculating the motion plan.
                  Defaults to ``None``.
                - ``"planner_id"``: (:obj:`str`) Name of the planner to use. Defaults to ``rrtconnect``.
                - ``"trials"``: (:obj:`int`) Number of search trials to run. Defaults to ``1``.
                - ``"resolution"``: (:obj:`float`) Validity checking resolution. This value
                  is specified as a fraction of the space's extent.
                  Defaults to ``0.02``.
                - ``"gantry_joint_limits"``: (:obj:`list` of `float`) List of 6 floats defining the upper/lower limits of
                  gantry joints. Use this if you want to restrict the working area of the path planner.
                - ``"arm_joint_limits"``: (:obj:`list` of `float`) List of 12 floats defining the upper/lower limits of
                  arm joints. Use this if you want to restrict the working area of the path planner.
                - ``"shallow_state_search"``: (:obj:`bool`) True to search only a minimum of
                  valid states before searching a path, False to search states intensively.
                - ``"optimize_path_length"``: (:obj:`bool`) True to search the path with minimal total length among all `trials`,
                  False to return the first valid path found. It only affects the output if `trials > 1`.
                - ``"log"``: (:class:`Logger`) Logger object to aid debugging. Default to ``None``.

        Returns:
            list: List of :class:`Configuration` objects representing the
            collision-free path to the ``goal_constraint``.
        """
        options = options or {}
        num_joints = len(robot.get_configurable_joints())
        metric_values = options.get('metric_values')
        collision_meshes = options.get('collision_meshes')
        planner_id = options.get('planner_id', 'rrtconnect')
        trials = options.get('trials', 1)
        resolution = options.get('resolution', 0.02)
        gantry_joint_limits = options.get('gantry_joint_limits')
        arm_joint_limits = options.get('arm_joint_limits')
        shallow_state_search = options.get('shallow_state_search', True)
        optimize_path_length = options.get('optimize_path_length', False)
        log = options.get('log')

        return self._find_path_plan(group, {'target_type': 'pose', 'target': goal_constraints},
                                    num_joints, metric_values, collision_meshes, planner_id, trials, resolution,
                                    gantry_joint_limits, arm_joint_limits, shallow_state_search, optimize_path_length,
                                    log)

    def plan_motion_to_config(self, robot, goal_configs, start_configuration=None, group=None, options=None):
        """Find a path plan to move the selected robot from its current position to one of the `goal_configs`.

        This function is useful when it is required to get a path plan that ends in one
        specific goal configuration.

        Args:
            robot (:class:`compas_fab.robots.Robot`): The robot instance for which the motion plan is being calculated.
            goal_configs (:obj:`list` of :class:`Configuration`): List of target or goal configurations.
            group (:obj:`int`): Integer referencing the desired robot group.
            start_configuration (:obj:`None`): Unused parameter.
            options (:obj:`dict`): Dictionary containing the following key-values pairs:

                - ``"metric_values"``: (:obj:`list` of :obj:`float`) List containing one value
                  per configurable joint. Each value ranges from 0 to 1,
                  where 1 indicates the axis/joint is blocked and cannot
                  move during inverse kinematic solving.
                - ``"collision_meshes"``: (:obj:`list` of :class:`compas.datastructures.Mesh`) Collision meshes
                  to be taken into account when calculating the motion plan.
                  Defaults to ``None``.
                - ``"planner_id"``: (:obj:`str`) Name of the planner to use. Defaults to ``rrtconnect``.
                - ``"trials"``: (:obj:`int`) Number of search trials to run. Defaults to ``1``.
                - ``"resolution"``: (:obj:`float`) Validity checking resolution. This value
                  is specified as a fraction of the space's extent.
                  Defaults to ``0.02``.
                - ``"gantry_joint_limits"``: (:obj:`list` of `float`) List of 6 floats defining the upper/lower limits of
                  gantry joints. Use this if you want to restrict the working area of the path planner.
                - ``"arm_joint_limits"``: (:obj:`list` of `float`) List of 12 floats defining the upper/lower limits of
                  arm joints. Use this if you want to restrict the working area of the path planner.
                - ``"shallow_state_search"``: (:obj:`bool`) True to search only a minimum of
                  valid states before searching a path, False to search states intensively.
                - ``"optimize_path_length"``: (:obj:`bool`) True to search the path with minimal total length among all `trials`,
                  False to return the first valid path found. It only affects the output if `trials > 1`.
                - ``"log"``: (:class:`Logger`) Logger object to aid debugging. Default to ``None``.

        Returns:
            list: List of :class:`Configuration` objects representing the
            collision-free path to the ``goal_configs``.
        """
        options = options or {}
        num_joints = len(robot.get_configurable_joints())
        metric_values = options.get('metric_values')
        collision_meshes = options.get('collision_meshes')
        planner_id = options.get('planner_id', 'rrtconnect')
        trials = options.get('trials', 1)
        resolution = options.get('resolution', 0.02)
        gantry_joint_limits = options.get('gantry_joint_limits')
        arm_joint_limits = options.get('arm_joint_limits')
        shallow_state_search = options.get('shallow_state_search', True)
        optimize_path_length = options.get('optimize_path_length', False)
        log = options.get('log')
        return self._find_path_plan(group, {'target_type': 'config', 'target': goal_configs},
                                    num_joints, metric_values, collision_meshes, planner_id, trials, resolution,
                                    gantry_joint_limits, arm_joint_limits, shallow_state_search, optimize_path_length,
                                    log)

    def _find_path_plan(self, group, goal, num_joints, metric_values, collision_meshes,
                        planner_id, trials, resolution,
                        gantry_joint_limits, arm_joint_limits, shallow_state_search, optimize_path_length,
                        log):
        if not metric_values:
            metric_values = [0.1] * num_joints

        if planner_id not in self.SUPPORTED_PLANNERS:
            raise ValueError('Unsupported planner_id. Must be one of: ' + str(self.SUPPORTED_PLANNERS))

        first_start = timer() if log else None
        if collision_meshes:
            self.client.add_collision_mesh(collision_meshes)
        if log:
            log.debug('Execution time: add_collision_mesh=%.2f', timer() - first_start)

        start = timer() if log else None
        self.client.set_robot_metric(group, metric_values)
        if log:
            log.debug('Execution time: set_robot_metric=%.2f', timer() - start)

        if 'target_type' not in goal:
            raise ValueError('Invalid goal type, you are using an internal function but passed incorrect args')

        if goal['target_type'] == 'config':
            states = []
            for c in goal['target']:
                states.extend(config_to_vrep(c, self.client.scale))
        elif goal['target_type'] == 'pose':
            start = timer() if log else None
            max_trials = None if shallow_state_search else 80
            max_results = 1 if shallow_state_search else 80
            states = self.client.find_raw_robot_states(group, frame_to_vrep_pose(goal['target'], self.client.scale), gantry_joint_limits, arm_joint_limits, max_trials, max_results)
            if log:
                log.debug('Execution time: search_robot_states=%.2f', timer() - start)

        start = timer() if log else None
        string_param_list = [planner_id]
        if gantry_joint_limits or arm_joint_limits:
            joint_limits = []
            joint_limits.extend(floats_to_vrep(gantry_joint_limits or [], self.client.scale))
            joint_limits.extend(arm_joint_limits or [])
            string_param_list.append(','.join(map(str, joint_limits)))

        if log:
            log.debug('About to execute path planner: planner_id=%s, trials=%d, shallow_state_search=%s, optimize_path_length=%s',
                      planner_id, trials, shallow_state_search, optimize_path_length)

        res, _, path, _, _ = self.client.run_child_script('searchRobotPath',
                                                          [group,
                                                           trials,
                                                           (int)(resolution * 1000),
                                                           1 if optimize_path_length else 0],
                                                          states, string_param_list)
        if log:
            log.debug('Execution time: search_robot_path=%.2f', timer() - start)

        if res != 0:
            raise VrepError('Failed to search robot path', res)

        if log:
            log.debug('Execution time: total=%.2f', timer() - first_start)

        return [config_from_vrep(path[i:i + num_joints], self.client.scale)
                for i in range(0, len(path), num_joints)]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.interfaces import ForwardKinematics
from compas_fab.backends.vrep.helpers import vrep_pose_to_frame

__all__ = [
    'VrepForwardKinematics',
]


class VrepForwardKinematics(ForwardKinematics):
    """Callable to calculate forward kinematics to get the current end-effector pose.
    """
    def __init__(self, client):
        self.client = client

    def forward_kinematics(self, robot=None, configuration=None, group=None, options=None):
        """Calculates forward kinematics to get the current end-effector pose.

        Args:
            robot (:class:`compas_fab.robots.Robot`): The robot instance for which forward kinematics is being calculated.
            configuration (:obj:`None`): Unused parameter.
            group (:obj:`int`): Integer referencing the desired robot group.
            options (:obj:`dict`): Unused parameter.

        Examples:

            >>> from compas_fab.robots import *
            >>> from compas_fab.backends import VrepClient
            >>> with VrepClient() as client:
            ...     frame = client.forward_kinematics(robot=None, configuration=None, group=0)

        Returns:
            An instance of :class:`Frame`.
        """

        _res, _, pose, _, _ = self.client.run_child_script('getIkTipPose', [group], [], [])
        return vrep_pose_to_frame(pose, self.client.scale)



"""This module is mostly for example purposes.

It needs to be explicitely imported, since it is not a fundamental part
of the compas_fab.robots package at all.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas.robots import LocalPackageMeshLoader
from compas.robots import RobotModel

import compas_fab
from compas_fab.robots import Robot as RobotClass
from compas_fab.robots import RobotSemantics

__all__ = [
    'Robot',
]


def Robot(client=None, load_geometry=False):
    """Returns a UR5 robot.

    This method serves mainly as help for writing examples, so that the code can
    stay short.

    It is intentionally capitalized to act as an almost drop-in replacement for
    the constructor of :class:`compas_fab.robots.Robot`.

    Parameters
    ----------
    client: object
        Backend client.
    load_geometry: bool

    Returns
    -------
    :class:`compas_fab.robots.Robot`
        Newly created instance of a UR5 robot.

    Examples
    --------

    >>> from compas_fab.robots.ur5 import Robot
    >>> robot = Robot()
    >>> robot.name
    'ur5'
    """

    urdf_filename = compas_fab.get('universal_robot/ur_description/urdf/ur5.urdf')
    srdf_filename = compas_fab.get('universal_robot/ur5_moveit_config/config/ur5.srdf')

    model = RobotModel.from_urdf_file(urdf_filename)
    semantics = RobotSemantics.from_srdf_file(srdf_filename, model)

    if load_geometry:
        loader = LocalPackageMeshLoader(compas_fab.get('universal_robot'), 'ur_description')
        model.load_geometry(loader)

    robot = RobotClass(model, semantics=semantics)

    if client:
        robot.client = client

    return robot



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import compas
import compas.plugins


@compas.plugins.plugin(category='install')
def installable_rhino_packages():
    return ['compas_fab', 'roslibpy']


@compas.plugins.plugin(category='install')
def after_rhino_install(installed_packages):
    if 'compas_fab' not in installed_packages:
        return []

    import compas_fab.ghpython.components.install
    return compas_fab.ghpython.components.install.install()


@compas.plugins.plugin(category='install')
def after_rhino_uninstall(installed_packages):
    if 'compas_fab' not in installed_packages:
        return []

    import compas_fab.ghpython.components.install
    return compas_fab.ghpython.components.install.uninstall()


if __name__ == "__main__":
    print('This installation method is obsolete.')
    print('Please use `python -m compas_rhino.install` instead')
    print('COMPAS FAB will be automatically detected and installed')



"""
********************************************************************************
compas_fab
********************************************************************************

.. currentmodule:: compas_fab

This library provides tools to plan and execute robotic fabrication processes using
the COMPAS Framework with a special focus on enabling its usage from within CAD
environments.

.. toctree::
    :maxdepth: 3

    compas_fab.backends
    compas_fab.robots
    compas_fab.utilities
    compas_fab.sensors
    compas_fab.blender
    compas_fab.ghpython
    compas_fab.rhino

"""

import os

from .__version__ import __author__
from .__version__ import __author_email__
from .__version__ import __copyright__
from .__version__ import __description__
from .__version__ import __license__
from .__version__ import __title__
from .__version__ import __url__
from .__version__ import __version__

HERE = os.path.dirname(__file__)
HOME = os.path.abspath(os.path.join(HERE, '../..'))
DATA = os.path.abspath(os.path.join(HERE, 'data'))


def _find_resource(filename):
    filename = filename.strip('/')
    return os.path.abspath(os.path.join(DATA, filename))


def get(filename):
    return _find_resource(filename)


# Check if COMPAS is installed from git
# If that's the case, try to append the current head's hash to __version__
try:
    git_head_file = os.path.abspath(os.path.join(HOME, '.git', 'HEAD'))

    if os.path.exists(git_head_file):
        # git head file contains one line that looks like this:
        # ref: refs/heads/main
        with open(git_head_file, 'r') as git_head:
            _, ref_path = git_head.read().strip().split(' ')
            ref_path = ref_path.split('/')

            git_head_refs_file = os.path.abspath(os.path.join(HOME, '.git', *ref_path))

        if os.path.exists(git_head_refs_file):
            with open(git_head_refs_file, 'r') as git_head_ref:
                git_commit = git_head_ref.read().strip()
                __version__ += '-' + git_commit[:8]
except Exception:
    pass

__all_plugins__ = ['compas_fab.rhino.install']
__all__ = ['__author__', '__author_email__', '__copyright__', '__description__', '__license__', '__title__', '__url__', '__version__', 'get']



from __future__ import print_function

import os

from compas.robots import RobotModel
from roslibpy import Message
from roslibpy import Param
from roslibpy import Ros
from roslibpy.actionlib import ActionClient
from roslibpy.actionlib import Goal

from compas_fab.backends.interfaces.client import ClientInterface
from compas_fab.backends.ros.exceptions import RosError
from compas_fab.backends.ros.fileserver_loader import RosFileServerLoader
from compas_fab.backends.ros.messages import ExecuteTrajectoryFeedback
from compas_fab.backends.ros.messages import ExecuteTrajectoryGoal
from compas_fab.backends.ros.messages import ExecuteTrajectoryResult
from compas_fab.backends.ros.messages import FollowJointTrajectoryFeedback
from compas_fab.backends.ros.messages import FollowJointTrajectoryGoal
from compas_fab.backends.ros.messages import FollowJointTrajectoryResult
from compas_fab.backends.ros.messages import Header
from compas_fab.backends.ros.messages import JointTrajectory as RosMsgJointTrajectory
from compas_fab.backends.ros.messages import JointTrajectoryPoint as RosMsgJointTrajectoryPoint
from compas_fab.backends.ros.messages import MoveItErrorCodes
from compas_fab.backends.ros.messages import RobotTrajectory
from compas_fab.backends.ros.messages import Time
from compas_fab.backends.ros.planner import MoveItPlanner
from compas_fab.backends.tasks import CancellableFutureResult
from compas_fab.robots import Robot
from compas_fab.robots import RobotSemantics

__all__ = [
    'RosClient',
]

PLANNER_BACKENDS = {
    'moveit': MoveItPlanner
}


class CancellableRosActionResult(CancellableFutureResult):
    def __init__(self, goal):
        super(CancellableRosActionResult, self).__init__()
        self.goal = goal

    def cancel(self):
        """Attempt to cancel the action.

        If the task is currently being executed and cannot be cancelled
        then the method will return ``False``, otherwise the call will
        be cancelled and the method will return ``True``.
        """
        if self.done:
            raise Exception('Already completed action cannot be cancelled')

        if self.goal.is_finished:
            return False

        self.goal.cancel()

        # NOTE: Check if we can output more meaning results than just "we tried to cancel"
        return True


class LocalCacheInfo(object):
    def __init__(self, use_local_cache, robot_name=None, local_cache_directory=None):
        self.use_local_cache = use_local_cache
        self.robot_name = robot_name
        self.local_cache_directory = local_cache_directory

    @classmethod
    def from_local_cache_directory(cls, local_cache_directory):
        """Construct local caching options based on the local cache directory.

        Based on the specified local cache directory, determine whether
        caching should be used at all, what the robot name is, and what the
        real caching directory to use is (which differs from the local cache
        directory specified to this function).

        Examples
        --------
        >>> info = LocalCacheInfo.from_local_cache_directory(None)
        >>> info.use_local_cache
        False


        >>> local_directory = os.path.join(os.path.expanduser('~'), 'robot_description', 'robocop')
        >>> info = LocalCacheInfo.from_local_cache_directory(local_directory)
        >>> info.use_local_cache
        True
        >>> info.robot_name
        'robocop'
        """
        if local_cache_directory is None:
            return LocalCacheInfo(use_local_cache=False)

        path_parts = local_cache_directory.rstrip(os.path.sep).split(os.path.sep)
        path_parts, robot_name = path_parts[:-1], path_parts[-1]
        local_cache_directory = os.path.sep.join(path_parts)

        return LocalCacheInfo(use_local_cache=True,
                              robot_name=robot_name,
                              local_cache_directory=local_cache_directory)


class RosClient(Ros, ClientInterface):
    """Interface to use ROS as backend via the **rosbridge**.

    The connection is managed by ``roslibpy``.

    :class:`.RosClient` is a context manager type, so it's best
    used in combination with the ``with`` statement to ensure
    resource deallocation.

    Parameters
    ----------
    host : :obj:`str`
        ROS bridge host. Defaults to ``localhost``.
    port : :obj:`int`
        Port of the ROS Bridge. Defaults to ``9090``.
    is_secure : :obj:`bool`
        ``True`` to indicate it should use a secure web socket, otherwise ``False``.
    planner_backend: str
        Name of the planner backend plugin to use. The plugin must be a sub-class of
        :class:`compas_fab.backends.PlannerInterface`. Defaults to ``"moveit"``,
        making use of :class:`compas_fab.backends.MoveItPlanner`.

    Examples
    --------

    >>> from compas_fab.backends import RosClient
    >>> with RosClient() as client:
    ...     print('Connected: %s' % client.is_connected)
    Connected: True

    Note
    ----
    For more examples, check out the :ref:`ROS examples page <ros_examples>`.
    """

    def __init__(self, host='localhost', port=9090, is_secure=False, planner_backend='moveit'):
        super(RosClient, self).__init__(host, port, is_secure)

        planner_backend_type = PLANNER_BACKENDS[planner_backend]
        self.planner = planner_backend_type(self)
        self._ros_distro = None

    def __enter__(self):
        self.run()

        return self

    def __exit__(self, *args):
        self.close()

    @property
    def ros_distro(self):
        """Retrieves the ROS version to which the client is connected (eg. kinetic)"""
        if not self._ros_distro:
            self._ros_distro = Param(self, '/rosdistro').get(timeout=1)

        return self._ros_distro

    def load_robot(self, load_geometry=False, urdf_param_name='/robot_description', srdf_param_name='/robot_description_semantic', precision=None, local_cache_directory=None):
        """Load an entire robot instance -including model and semantics- directly from ROS.

        Parameters
        ----------
        load_geometry : bool, optional
            ``True`` to load the robot's geometry, otherwise ``False`` to load only the model and semantics.
        urdf_param_name : str, optional
            Parameter name where the URDF is defined. If not defined, it will default to ``/robot_description``.
        srdf_param_name : str, optional
            Parameter name where the SRDF is defined. If not defined, it will default to ``/robot_description_semantic``.
        precision : float
            Defines precision for importing/loading meshes. Defaults to ``compas.PRECISION``.
        local_cache_directory : str, optional
            Directory where the robot description (URDF, SRDF and meshes) are stored.
            This differs from the directory taken as parameter by the :class:`RosFileServerLoader`
            in that it points directly to the specific robot package, not to a global workspace storage
            for all robots. If not assigned, the robot will not be cached locally.

        Examples
        --------

        >>> from compas_fab.backends import RosClient
        >>> with RosClient() as client:
        ...     robot = client.load_robot()
        ...     print(robot.name)
        ur5
        """
        robot_name = None

        cache_info = LocalCacheInfo.from_local_cache_directory(local_cache_directory)
        use_local_cache = cache_info.use_local_cache
        robot_name = cache_info.robot_name
        local_cache_directory = cache_info.local_cache_directory

        loader = RosFileServerLoader(self, use_local_cache, local_cache_directory, precision)

        if robot_name:
            loader.robot_name = robot_name

        urdf = loader.load_urdf(urdf_param_name)
        srdf = loader.load_srdf(srdf_param_name)

        model = RobotModel.from_urdf_string(urdf)
        semantics = RobotSemantics.from_srdf_string(srdf, model)

        if load_geometry:
            model.load_geometry(loader)

        return Robot(model, semantics=semantics, client=self)

    # ==========================================================================
    # executing
    # ==========================================================================

    def get_configuration(self):
        pass

    def follow_configurations(self, callback, joint_names, configurations, timesteps, timeout=60000):

        if len(configurations) != len(timesteps):
            raise ValueError("%d configurations must have %d timesteps, but %d given." % (
                len(configurations), len(timesteps), len(timesteps)))

        if not timeout:
            timeout = timesteps[-1] * 1000 * 2

        points = []
        num_joints = len(configurations[0].joint_values)
        for config, time in zip(configurations, timesteps):
            pt = RosMsgJointTrajectoryPoint(positions=config.joint_values, velocities=[
                                      0] * num_joints, time_from_start=Time(secs=(time)))
            points.append(pt)

        joint_trajectory = RosMsgJointTrajectory(
            Header(), joint_names, points)  # specify header necessary?
        return self.follow_joint_trajectory(joint_trajectory=joint_trajectory, callback=callback, timeout=timeout)

    def _convert_to_ros_trajectory(self, joint_trajectory):
        """Converts a ``compas_fab.robots.JointTrajectory`` into a ROS Msg joint trajectory."""

        # For backwards-compatibility, accept ROS Msg directly as well and simply do not modify
        if isinstance(joint_trajectory, RosMsgJointTrajectory):
            return joint_trajectory

        trajectory = RosMsgJointTrajectory()
        trajectory.joint_names = joint_trajectory.joint_names

        for point in joint_trajectory.points:
            ros_point = RosMsgJointTrajectoryPoint(
                positions=point.positions,
                velocities=point.velocities,
                accelerations=point.accelerations,
                effort=point.effort,
                time_from_start=Time(
                    point.time_from_start.secs, point.time_from_start.nsecs),
            )
            trajectory.points.append(ros_point)

        return trajectory

    def follow_joint_trajectory(self, joint_trajectory, action_name='/joint_trajectory_action', callback=None, errback=None, feedback_callback=None, timeout=60000):
        """Follow the joint trajectory as computed by MoveIt planner.

        Parameters
        ----------
        joint_trajectory : :class:`compas_fab.robots.JointTrajectory`
            Instance of joint trajectory. Note: for backwards compatibility, this supports a ROS Msg being passed as well.
        action_name : string
            ROS action name, defaults to ``/joint_trajectory_action`` but some drivers need ``/follow_joint_trajectory``.
        callback : callable
            Function to be invoked when the goal is completed, requires
            one positional parameter ``result``.
        errback : callable
            Function to be invoked in case of error or timeout, requires
            one position parameter ``exception``.
        feedback_callback : callable
            Function to be invoked during execution to provide feedback.
        timeout : int
            Timeout for goal completion in milliseconds.

        Returns
        -------
        :class:`CancellableTask`
            An instance of a cancellable tasks.
        """

        joint_trajectory = self._convert_to_ros_trajectory(joint_trajectory)
        trajectory_goal = FollowJointTrajectoryGoal(trajectory=joint_trajectory)

        action_client = ActionClient(self, action_name, 'control_msgs/FollowJointTrajectoryAction')
        goal = Goal(action_client, Message(trajectory_goal.msg))
        action_result = CancellableRosActionResult(goal)

        def handle_result(msg):
            result = FollowJointTrajectoryResult.from_msg(msg)
            if result.error_code != FollowJointTrajectoryResult.SUCCESSFUL:
                ros_error = RosError(
                    'Follow trajectory failed={}'.format(result.error_string),
                    int(result.error_code))
                action_result._set_error_result(ros_error)
                if errback:
                    errback(ros_error)
            else:
                action_result._set_result(result)
                if callback:
                    callback(result)

        def handle_feedback(msg):
            feedback = FollowJointTrajectoryFeedback.from_msg(msg)
            feedback_callback(feedback)

        def handle_failure(error):
            errback(error)

        goal.on('result', handle_result)

        if feedback_callback:
            goal.on('feedback', handle_feedback)

        if errback:
            goal.on('timeout', lambda: handle_failure(
                RosError('Action Goal timeout', -1)))

        goal.send(timeout=timeout)

        return action_result

    def execute_joint_trajectory(self, joint_trajectory, action_name='/execute_trajectory', callback=None, errback=None, feedback_callback=None, timeout=60000):
        """Execute a joint trajectory via the MoveIt infrastructure.

        Note
        ----
        This method does not support Multi-DOF Joint Trajectories.

        Parameters
        ----------
        joint_trajectory : :class:`compas_fab.robots.JointTrajectory`
            Instance of joint trajectory.
        callback : callable
            Function to be invoked when the goal is completed, requires
            one positional parameter ``result``.
        action_name : string
            ROS action name, defaults to ``/execute_trajectory``.
        errback : callable
            Function to be invoked in case of error or timeout, requires
            one position parameter ``exception``.
        feedback_callback : callable
            Function to be invoked during execution to provide feedback.
        timeout : int
            Timeout for goal completion in milliseconds.

        Returns
        -------
        :class:`CancellableFutureResult`
            An instance of a cancellable future result.
        """

        joint_trajectory = self._convert_to_ros_trajectory(joint_trajectory)
        trajectory = RobotTrajectory(joint_trajectory=joint_trajectory)
        trajectory_goal = ExecuteTrajectoryGoal(trajectory=trajectory)

        action_client = ActionClient(self, action_name, 'moveit_msgs/ExecuteTrajectoryAction')
        goal = Goal(action_client, Message(trajectory_goal.msg))
        action_result = CancellableRosActionResult(goal)

        def handle_result(msg):
            result = ExecuteTrajectoryResult.from_msg(msg)
            if result.error_code != MoveItErrorCodes.SUCCESS:
                ros_error = RosError('Execute trajectory failed. Message={}'.format(result.error_code.human_readable), int(result.error_code))
                action_result._set_error_result(ros_error)
                if errback:
                    errback(ros_error)
            else:
                action_result._set_result(result)
                if callback:
                    callback(result)

        def handle_feedback(msg):
            feedback = ExecuteTrajectoryFeedback.from_msg(msg)
            feedback_callback(feedback)

        def handle_failure(error):
            errback(error)

        goal.on('result', handle_result)

        if feedback_callback:
            goal.on('feedback', handle_feedback)

        if errback:
            goal.on('timeout', lambda: handle_failure(RosError('Action Goal timeout', -1)))

        goal.send(timeout=timeout)

        return action_result



import time

from compas_fab.sensors import SerialSensor
from compas_fab.sensors.exceptions import ProtocolError
from compas_fab.sensors.exceptions import SensorTimeoutError

__all__ = ['PosCon3D', 'PosConCM']


class PosCon3D(SerialSensor):
    """Provides an interface for the Baumer PosCon3D edge measurement sensor.

    The sensor has different interfaces to retrieve its data. This
    class provides access to the serial interface (RS-485). This
    class is a context manager type, so it's best used in combination
    with the ``with`` statement to ensure resource deallocation. The
    protocol of the sensor when operated via RS-485 indicates that
    access to it must be locked programmatically before starting operations
    and unlocked on completion. This is handled automatically if you use
    this class on a ``with`` statement, otherwise, the methods ``begin()`` and
    ``end()`` must be invoked by hand.

    Attributes
    ----------
    serial : :obj:`serial.Serial`
        Instance of the serial port used to communicate with the sensor.
    address : :obj:`int`
        PosCon3D sensors have an address assigned, which defaults to 1.
        There's also a broadcast address (``PosCon3D.BROADCAST_ADDRESS``)
        that can be used to query the address of the sensor connected to
        the RS-485 bus. Only one sensor can be in the bus when using the
        broadcast address to query for sensor's address.

    Examples
    --------
    >>> from serial import Serial                                                   # doctest: +SKIP
    >>> with Serial('COM5', 57600, timeout=1) as serial:                            # doctest: +SKIP
    >>>     with PosCon3D(serial, PosCon3D.BROADCAST_ADDRESS) as broadcast_query:   # doctest: +SKIP
    ...         addr = broadcast_query.get_address()                                # doctest: +SKIP
    ...                                                                             # doctest: +SKIP
    ...     with PosCon3D(serial, addr) as sensor:                                  # doctest: +SKIP
    ...         sensor.set_measurement_type('Edge L rise')                          # doctest: +SKIP
    ...         sensor.set_precision(2)                                             # doctest: +SKIP
    ...         data = sensor.get_measurement()                                     # doctest: +SKIP
    """
    FRAME_HEAD = '{%s,%s,%s'
    FRAME_TAIL = '%s%s}'
    BROADCAST_ADDRESS = 0
    MEASUREMENT_TYPES = ('Edge L rise', 'Edge L fall', 'Edge R rise',
                         'Edge R fall', 'Width', 'Center width', 'Gap', 'Center gap')
    QUALITY = {
        0: 'Valid',
        1: 'Low signal',
        2: 'No edge',
        3: 'Low signal, no edge',
        4: 'No signal'}
    ERROR_CODES = {
        '000': 'No error',
        '001': 'False checksum',
        '002': 'False command',
        '003': 'False frame',
        '004': 'False value or parameter',
        '005': 'Missed command 000 to begin RS-485 control',
        '006': 'Out of range',
        '007': 'Buffer overflow',
        '010': 'All outputs Off',
        '020': 'Display Off',
        '99': 'Argument out of Range',
        '100': 'Distance out of Range (see FSP)',
        '101': 'Angle out of Range (see FSP)',
        '102': 'Flatness out of Range (see FSP)',
        '103': 'Length out of Range (see FSP)',
        '200': 'Fatal Error (Reset sensor, Power Off / On)'}

    def __init__(self, serial, address):
        super(PosCon3D, self).__init__(serial)
        self.address = address

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, *args):
        self.end()

    def begin(self):
        """Locks the sensor to start RS-485 communication.

        Note
        ----
        This method only needs to be called if not using
        a ``with`` statement to handle lifetime of the `PosCon3D` instance.
        """
        return self.send_command(self.address, '000', '1')

    def end(self):
        """Unlocks the sensor from RS-485 communication.

        Note
        ----
        This method only needs to be called if not using
        a ``with`` statement to handle lifetime of the `PosCon3D` instance.
        """
        return self.send_command(self.address, '000', '0')

    def send_command(self, address, command, data=None):
        """Sends a command to the sensor's address specified. The command
        can optionally contain a data string.

        This method is mostly for internal use, as the higher-level API is
        exposed via dedicated methods.

        Parameters
        ----------
        address : :obj:`int`
            PosCon3D sensors have an address assigned, which defaults to 1.
            There's also a broadcast address (``PosCon3D.BROADCAST_ADDRESS``)
            that can be used to query the address of the sensor connected to
            the RS-485 bus. Only one sensor can be in the bus when using the
            broadcast address to query for sensor's address.
        command : :obj:`string`
            A string indicating the command number to be executed.
        data : :obj:`string`
            An optional string of data that is sent together with
            the command.

        Returns
        -------
        `list` or value
            Result of the command. It can be a list or a single value depending on the operation.
        """
        cmd = self.format_command(address, command, data)

        # Python 2 vs 3
        if hasattr(cmd, 'decode'):
            self.serial.write(cmd)
            result = self.serial.readline()
        else:
            self.serial.write(cmd.encode('ascii'))
            result = self.serial.readline().decode('ascii')

        if result:
            frame_head = result[:-4]
            checksum = result[-4:-1]
            expected = self.calculate_checksum(frame_head)
            if expected != checksum:
                raise ProtocolError('Invalid response, checksum mismatch. Expected=%s, Got=%s' % (expected, checksum))

            expected_frame_head = self.FRAME_HEAD % (address, command, '')
            if not result.startswith(expected_frame_head):
                raise ProtocolError('Invalid response, command/address mismatch. Expected to start with="%s", Got="%s"' % (expected_frame_head, result))

            return self.get_payload(result)

        return None

    def format_command(self, address, command, data=None):
        """Formats the command."""
        data = data + ',' if data else ''
        frame = self.FRAME_HEAD % (address, command, data)
        return self.FRAME_TAIL % (frame, self.calculate_checksum(frame))

    def calculate_checksum(self, command):
        """Checks that message is complete."""
        code_points = [ord(c) for c in list(command)]
        checksum = 0
        for i in code_points:
            checksum = checksum ^ i

        return str(checksum).zfill(3)

    def get_payload(self, result):
        """Gets payload."""
        data = result.split(',')[2:-1]
        if not data:
            return None
        elif len(data) == 1:
            return data[0]
        else:
            if data[0] == 'E':
                raise ProtocolError(self.ERROR_CODES[str(data[1])])

            return data

    def get_address(self):
        """Gets the address of the RS-485 sensors currently connected to the bus. This command
        is only really useful when this class is initialized with the broadcast address,
        with the purpose of retrieving the address of a sensor connected.

        Returns
        -------
        `int`
            Address of the PosCon3D sensor connected to the RS-485 bus.

        Note
        ----
        Only one PosCon3D sensor can be connected to the bus for this operation to succeed.
        """
        return int(self.send_command(self.address, '013'))

    def set_measurement_type(self, measurement_type):
        """Defines the measurement type to use.

        =================== =========== ======
        Measurement type    Function    Value
        =================== =========== ======
        "Edge L rise"       Edge        0
        "Edge L fall"       Edge        1
        "Edge R rise"       Edge        2
        "Edge R fall"       Edge        3
        "Width"             Width       4
        "Center width"      Width       5
        "Gap"               Gap         6
        "Center gap"        Gap         7
        =================== =========== ======

        Parameters
        ----------
        measurement_type : :obj:`string`
            Measurement type.

        """
        if measurement_type not in self.MEASUREMENT_TYPES:
            raise ProtocolError('Unsupported measure type, must be one of ' + str(self.MEASUREMENT_TYPES))

        return self.send_command(self.address, '020', str(self.MEASUREMENT_TYPES.index(measurement_type)))

    def set_precision(self, precision):
        """Defines the precision the sensor will use to determine edges:

        ======== ============ ===============
        Value    Precision    Function values
        ======== ============ ===============
        ``0``    Standard     Median=off, Moving Average=off
        ``1``    High         Median=7, Moving Average=16
        ``2``    Very High    Median=15, Moving Average=128
        ======== ============ ===============

        Parameters
        ----------
        precision : :obj:`int`
            Sensor precision to use.

        Note
        ----
        The higher the precision, the slower the measurement gets.
        """
        if precision < 0 or precision > 2:
            raise ProtocolError('Precision must be 0 (standard), 1 (high) or 2 (very high)')
        return self.send_command(self.address, '040', str(precision))

    def set_edge_height(self, height):
        """Defines the minimum height of an edge to be detected.

        Parameters
        ----------
        height : :obj:`float`
            Minimum edge height.
        """
        return self.send_command(self.address, '042', str(height))

    def get_measurement(self):
        """Retrieves the current measurement of the sensor according to the current settings.

        Returns
        -------
        `tuple`
            The current measurement and additionally a value indicating the quality of the measured value.
        """
        result = self.send_command(self.address, '031')

        if len(result) != 2:
            raise ProtocolError('Unexpected result: ' + str(result))

        value = result[0]
        quality = int(result[1])

        # If Valid
        if quality == 0:
            value = float(value)

        return (value, self.QUALITY[quality])

    def get_live_monitor_data(self):
        """Retrieves the distance to the surface in the center of the laser beam and the
        angle at which it's found.

        Returns
        -------
        `list`
            angle and distance to the reference surface.

        Note
        ----
        This function is designed to aid in the installation of the sensor at an angle.
        """
        result = self.send_command(self.address, '093')
        if len(result) != 2:
            raise ProtocolError('Unexpected result: ' + str(result))

        return map(float, result)

    def reset(self):
        """Resets the sensor to factory settings."""
        self.send_command(self.address, '003')

    def activate_flex_mount(self, reference_thickness):
        """Activates the FLEX Mount feature of the sensor to allow positioning it on an
        angled installation. The reference thickness is only required if the surface is
        uneven and an additional leveling auxiliary plate as been added."""
        result = self.send_command(self.address, '062', str(reference_thickness))
        return map(float, result)

    def deactivate_flex_mount(self):
        """Deactivates the FLEX Mount feature."""
        self.send_command(self.address, '063')

    def set_flex_mount(self, angle, distance):
        """Sets the FLEX Mount feature to a specific angle and distance."""
        result = self.send_command(self.address, '060', '%.2f,%.2f' % (angle, distance))
        return map(float, result)

    def adjust_to_dark_object(self, is_dark_object):
        """Adjusts the sensor to detect darker or lighter surfaces."""
        data = '1' if is_dark_object else '0'
        return self.send_command(self.address, '044', data)


class PosConCM(SerialSensor):
    """Provides an interface for the Baumer PosConCM round objects measurement sensor.

    The sensor has different interfaces to retrieve its data. This
    class provides access to the serial interface (RS-485).
    This class is a context manager type, so it's best used in combination
    with the ``with`` statement to ensure resource deallocation.
    The protocol of the sensor when operated via RS-485 indicates that
    access to it must be locked programmatically before starting operations
    and unlocked on completion. This is handled automatically if you use
    this class on a ``with`` statement, otherwise, the methods ``begin()`` and
    ``end()`` must be invoked by hand.

    Attributes
    ----------
    serial : :obj:`serial.Serial`
        Instance of the serial port used to communicate with the sensor.
    address : :obj:`int`
        PosConCM sensors have an address assigned, which defaults to 1.
        There's also a broadcast address (``PosConCM.BROADCAST_ADDRESS``)
        that can be used to query the address of the sensor connected to
        the RS-485 bus. Only one sensor can be in the bus when using the
        broadcast address to query for sensor's address.

    Examples
    --------
    >>> from serial import Serial                                                   # doctest: +SKIP
    >>> with Serial('COM5', 57600, parity=PARITY_EVEN, timeout=1) as serial:                            # doctest: +SKIP
    >>>     with PosConCM(serial, PosConCM.BROADCAST_ADDRESS) as broadcast_query:   # doctest: +SKIP
    ...         addr = broadcast_query.get_address()                                # doctest: +SKIP
    ...                                                                             # doctest: +SKIP
    ...     with PosConCM(serial, addr) as sensor:                                  # doctest: +SKIP
    ...         sensor.set_measurement_type('X-Center')                             # doctest: +SKIP
    ...         sensor.set_precision(2)                                             # doctest: +SKIP
    ...         data = sensor.get_measurement()                                     # doctest: +SKIP
    """

    FRAME_HEAD = ':%s%s;%s;'
    FRAME_TAIL = '%s%s\r\n'
    BROADCAST_ADDRESS = 0
    MEASUREMENT_TYPES = {
        'Diameter': 28,
        'X_center': 29,
        'Z_center': 30,
        'X_left': 31,
        'X_right': 32,
        'Z_top': 33}
    QUALITY = {
        0: 'Valid',
        1: 'Low signal',
        2: 'No edge',
        3: 'Low signal, no edge',
        4: 'No signal'}
    ERROR_CODES = {
        '1': 'Wrong message type',
        '2': 'Wrong payload format',
        '3': 'Wrong argument',
        '4': 'Wrong argument count',
        '5': 'Not enough data',
        '6': 'Index do not exist',
        '7': 'Index locked',
        '8': 'Access not allowed',
        '9': 'Not enough memory for encoding',
        '10': 'Not possible to encode argument',
        '11': 'Application specific error',
        '12': 'Wrong state'}

    def __init__(self, serial, address):
        super(PosConCM, self).__init__(serial)
        self.address = address

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, *args):
        self.end()

    def begin(self):
        """Locks the sensor to start RS-485 communication.

        Note
        ----
        This method only needs to be called if not using
        a ``with`` statement to handle lifetime of the `PosConCM` instance.
        """
        return self.send_command(self.address, 'W010', '0')

    def end(self):
        """Unlocks the sensor from RS-485 communication.

        Note
        ----
        This method only needs to be called if not using
        a ``with`` statement to handle lifetime of the `PosConCM` instance.
        """
        return self.send_command(self.address, 'W010', '1')

    def format_command(self, address, command, data=None):
        """Formats the command."""
        data = data or ''
        frame = self.FRAME_HEAD % (str(address).zfill(2), command, data)
        return self.FRAME_TAIL % (frame, self.calculate_checksum(frame))

    def calculate_checksum(self, command):
        """Checks that message is complete."""
        return '****'

    def get_payload(self, result):
        """Gets payload."""
        frame_head = result[:-6]
        result_type = frame_head[3]

        if result_type == 'a':
            raise SensorTimeoutError('Sensor has not completed reading')

        if result_type == 'E':
            error_index = frame_head.split(';')
            raise ProtocolError('Application error, Result=%s' % frame_head +
                                'Error type: ' + str(self.ERROR_CODES[str(error_index[1])]))

        if result_type == 'B':
            raise ProtocolError('Sensor is busy, Result=%s' % frame_head)

        return result[5:-6].split(';')

    def send_command(self, address, command, data=None):
        """Sends a command to the sensor's address specified. The command
        can optionally contain a data string.
        This method is mostly for internal use, as the higher-level API is
        exposed via dedicated methods.

        Parameters
        ----------
        address : :obj:`int`
            PosConCM sensors have an address assigned, which defaults to 1.
            There's also a broadcast address (``PosConCM.BROADCAST_ADDRESS``)
            that can be used to query the address of the sensor connected to
            the RS-485 bus. Only one sensor can be in the bus when using the
            broadcast address to query for sensor's address.
        command : :obj:`string`
            A string indicating the command number to be executed with the
            W or R in front, depending on Writting or Reading.
        data : :obj:`string`
            An optional string of data that is sent together with the command.

        Returns
        -------
        `list` or value
            Result of the command. It can be a list or a single value depending on the operation.
        """
        for i in range(2):
            cmd = self.format_command(address, command, data)

            # Python 2 vs 3
            if hasattr(cmd, 'decode'):
                self.serial.write(cmd)
                result = self.serial.readline()
            else:
                self.serial.write(cmd.encode('ascii'))
                result = self.serial.readline().decode('ascii')

            if result:
                try:
                    return self.get_payload(result)
                except SensorTimeoutError:
                    time.sleep(0.5)
                    continue

            return None

    def get_address(self):
        """Gets the address of the RS-485 sensors currently connected to the bus. This command
        is only really useful when this class is initialized with the broadcast address,
        with the purpose of retrieving the address of a sensor connected.

        Returns
        -------
        `int`
            Address of the PosConCM sensor connected to the RS-485 bus.

        Note
        ----
        Only one PosConCM sensor can be connected to the bus for this operation to succeed.
        """
        result = self.send_command(self.address, 'R005')
        return int(result[0])

    def set_measurement_type(self, measurement_type):
        """Defines the measurement type to use.

        ===================  ======
        Measurement type     Value
        ===================  ======
        "Diameter"           28
        "X-Center position"  29
        "Z-Center position"  30
        "X-Left position"    31
        "X-Right position"   32
        "Z-Top position"     33
        ===================  ======

        Parameters
        ----------
        measurement_type : :obj:`string`
            Measurement type.
        """
        if measurement_type not in self.MEASUREMENT_TYPES:
            raise ProtocolError('Unsupported measure type, must be one of ' + str(self.MEASUREMENT_TYPES))

        return self.send_command(self.address, 'W020', str(self.MEASUREMENT_TYPES[measurement_type]))

    def set_precision(self, precision):
        """Defines the precision the sensor will use to determine edges:

        ======== ============ ===============================
        Value    Precision    Function values
        ======== ============ ===============================
        ``0``    Standard     Median=off,  Moving Average=off
        ``1``    High         Median=7,    Moving Average=16
        ``2``    Very High    Median=15,   Moving Average=128
        ======== ============ ===============================

        Parameters
        ----------
        precision : :obj:`int`
            Sensor precision to use.

        Note
        -----
        The higher the precision, the slower the measurement gets.
        """
        if precision < 0 or precision > 2:
            raise ProtocolError('Precision must be 0 (standard), 1 (high) or 2 (very high)')
        return self.send_command(self.address, 'W033', str(precision))

    def get_measurement(self):
        """Retrieves the current measurement of the sensor according to the current settings.

        Returns
        -------
        `tuple`
            The current measurement and additionally a value indicating the quality of the measured value.
        """
        result = self.send_command(self.address, 'R021')

        if len(result) != 3:
            raise ProtocolError('Unexpected result: ' + str(result))

        value = result[0]
        quality = int(result[1])

        # If Valid
        if quality == 0:
            value = float(value)

        return (value, self.QUALITY[quality])

    def activate_flex_mount(self):
        """Activates the FLEX Mount feature of the sensor to allow positioning it on an
        angled installation. The reference thickness is only required if the surface is
        uneven and an additional leveling auxiliary plate as been added."""
        result = self.send_command(self.address, 'W035', '1')
        return map(float, result)

    def deactivate_flex_mount(self):
        """Deactivates the FLEX Mount feature."""
        result = self.send_command(self.address, 'W035', '0')
        return map(float, result)

    def set_flex_mount(self, angle, distance):
        """Sets the FLEX Mount feature to a specific angle and distance."""
        data = '{:.2f};{:.2f}'.format(angle, distance)
        return self.send_command(self.address, 'W036', data)

    def teach_flex_mount(self, reference_thickness):
        """Sets the FLEX Mount feature to a specific angle and distance."""
        return self.send_command(self.address, 'W037', str(reference_thickness))

    def adjust_to_dark_object(self, is_dark_object):
        """Adjusts the sensor to detect darker or lighter surfaces."""
        data = '1' if is_dark_object else '0'
        result = self.send_command(self.address, 'W032', data)
        return map(float, result)

    def reset(self):
        """Resets the sensor to factory settings."""
        self.send_command(self.address, 'W202', '0')
